import copy
import warnings
from typing import Union, List

from qualang_tools.config.components import (
    Controller,
    ElementCollection,
    Element,
    Mixer,
    Oscillator,
    MeasureElement,
    MeasurePulse,
)
from qualang_tools.config.configuration import QMConfiguration
from qualang_tools.config.exceptions import ConfigurationError
from qualang_tools.config.parameters import _is_parametric
from qualang_tools.config.primitive_components import (
    AnalogOutputPort,
    AnalogInputPort,
    DigitalOutputPort,
    Waveform,
    Pulse,
    IntegrationWeights,
    DigitalInputPort,
    Port,
)


class ConfigBuilder:
    def __init__(self) -> None:
        """A class used to build the configuration of an experiment with QUA"""
        warnings.warn(
            "ConfigBuilder is no longer being actively developed and may not support all config functionality",
            DeprecationWarning,
            stacklevel=2,
        )
        self.objects = []
        self.configuration = QMConfiguration([])
        self.controllers = []
        self.components: List[ElementCollection] = []

    def add(self, obj):
        """Adds an entity to the configuraiton.

        :param obj: Any entity type supported by the configuration system, e.g. Element, ArbitraryWaveform or Transmon
        :type obj:
        """
        if isinstance(obj, list):
            for _obj in obj:
                if isinstance(_obj, ElementCollection):
                    self.components.append(_obj)
                else:
                    self.objects.append(_obj)
        else:
            if isinstance(obj, ElementCollection):
                self.components.append(obj)
            else:
                self.objects.append(obj)
        return self

    @property
    def ports(self):
        ports = []
        for cont in self.controllers:
            ports += cont.ports
        return ports

    def _add_controller(self, cont: Controller):
        self.controllers.append(cont)
        for port in cont.ports:
            port_info = port.info[1]
            if isinstance(port, AnalogOutputPort):
                port_dict = cont.dict["analog_outputs"][port_info]
                if port.delay is not None:
                    port_dict["delay"] = port.delay
                if port.filter is not None:
                    port_dict["filter"] = {
                        "feedback": port.filter.feedback,
                        "feedforward": port.filter.feedforward,
                    }
                if port.crosstalk is not None:
                    port_dict["crosstalk"] = port.crosstalk
                port_dict["offset"] = port.offset
                port_dict["shareable"] = port.shareable
            elif isinstance(port, AnalogInputPort):
                port_dict = cont.dict["analog_inputs"][port_info]
                if port.gain_db is not None:
                    port_dict["gain_db"] = port.gain_db
                port_dict["offset"] = port.offset
                port_dict["shareable"] = port.shareable
            elif isinstance(port, DigitalOutputPort):
                port_dict = cont.dict["digital_outputs"][port_info]
                port_dict["shareable"] = port.shareable
            elif isinstance(port, DigitalInputPort):
                port_dict = cont.dict["digital_inputs"][port_info]
                if port.polarity is not None:
                    port_dict["polarity"] = port.polarity
                if port.threshold is not None:
                    port_dict["threshold"] = port.threshold
                if port.window is not None:
                    port_dict["window"] = port.window
                port_dict["shareable"] = port.shareable
        self.configuration.config["controllers"][cont.name] = cont.dict

    def build(self, config=None):
        """Generates the QUA configuration represented by the current state of the builder object

        :param config: An initial QUA config, defaults to {}
        :type config: Dict, optional
        :return: QUA configuration
        :rtype: Dict
        """
        if config is None:
            config = {}
        if config == {}:
            self.configuration.reset()
            self.controllers = []
        else:
            self.configuration.config = copy.deepcopy(config)
        objects = self.objects
        for c in self.components:
            objects = objects + c.get_elements()

        for obj in objects:
            if _is_parametric(obj):
                raise ConfigurationError("set the parameters of {0}".format(obj.name))
            if isinstance(obj, Controller):
                self._add_controller(obj)
            # elif isinstance(obj, ElementCollection):
            #    for _obj in obj.get_elements():
            #        self._objects.append(_obj)
            elif isinstance(obj, Waveform):
                self.configuration.add_waveform(obj)
            elif isinstance(obj, Pulse):
                self.configuration.add_pulse(obj)
            elif isinstance(obj, Element):
                self.configuration.add_element(obj)
            elif isinstance(obj, IntegrationWeights):
                self.configuration.add_integration_weights(obj)
            elif isinstance(obj, Mixer):
                self.configuration.add_mixer(obj)
            elif isinstance(obj, Oscillator):
                self.configuration.add_oscillator(obj)
            else:
                raise NotImplementedError("Can not add objects of type {0}".format(type(obj)))
        return self.configuration.config

    def find_instances(self, obj_type):
        """Find instances of a certain class in the configuration

        :param obj_type: The required type
        :type obj_type: [type]
        :return: A list of instances
        :rtype: list
        """
        return [elm for elm in self.objects if isinstance(elm, obj_type)]

    def find_users_of(
        self,
        elm: Union[
            AnalogInputPort,
            AnalogOutputPort,
            Controller,
            Waveform,
            Pulse,
            Mixer,
            Oscillator,
        ],
    ):
        """Find where in the configuration a specific object is used

        :param elm: The element of interest
        :type elm: Union[AnalogInputPort, AnalogOutputPort, Controller, Waveform, Pulse]
        :return: A set of users
        :rtype: Set
        """
        if (
            isinstance(elm, AnalogInputPort)
            or isinstance(elm, AnalogOutputPort)
            or isinstance(elm, DigitalInputPort)
            or isinstance(elm, DigitalOutputPort)
        ):
            return self._find_users_of_port(elm)
        elif isinstance(elm, Controller):
            return self._find_users_of_controller(elm)
        elif isinstance(elm, Waveform):
            return self._find_users_of_waveform(elm)
        elif isinstance(elm, Pulse):
            return self._find_users_of_pulse(elm)
        elif isinstance(elm, IntegrationWeights):
            return self._find_users_of_integration_weights(elm)
        elif isinstance(elm, Mixer):
            return self._find_users_of_mixer(elm)
        else:
            raise NotImplementedError("can not find objects of type {}".format(type(elm)))

    def _find_users_of_port(self, port: Port):
        objects = []
        for obj in self.objects:
            if isinstance(obj, Element) or isinstance(obj, MeasureElement):
                if port in obj.ports:
                    objects.append(obj)
        return set(objects)

    def _find_users_of_controller(self, cont: Controller):
        objects = []
        for obj in self.objects:
            if isinstance(obj, Element) or isinstance(obj, MeasureElement):
                for port in obj.ports:
                    if port in cont.ports:
                        objects.append(obj)
                        break
        return set(objects)

    def _find_users_of_waveform(self, wf: Waveform):
        objects = []
        for obj in self.objects:
            if isinstance(obj, Element) or isinstance(obj, MeasureElement):
                for name in obj.waveform_names:
                    if name == wf.name:
                        objects.append(obj)
        return set(objects)

    def _find_users_of_pulse(self, elm: Pulse):
        objects = []
        for obj in self.objects:
            if isinstance(obj, Element) or isinstance(obj, MeasureElement):
                for name in obj.pulse_names:
                    if name == elm.name:
                        objects.append(obj)
        return set(objects)

    def _find_users_of_integration_weights(self, elm: IntegrationWeights):
        objects = []
        for obj in self.objects:
            if isinstance(obj, MeasurePulse):
                for w in obj.integration_weights:
                    if w.name == elm.name:
                        objects.append(obj)
        return set(objects)

    def _find_users_of_mixer(self, elm: Mixer):
        objects = []
        for obj in self.objects:
            if isinstance(obj, Oscillator):
                if obj.mixer == elm.name:
                    objects.append(obj)
            elif isinstance(obj, Element):
                if obj.mixer.name == elm.name:
                    objects.append(obj)
        return set(objects)

    def component(self, name: str):
        """Returns a config builder object with the given name
        :param name: name of the object
        :type name: str
        """
        for c in self.objects + self.components:
            if c.name == name:
                return c



from ._cirq import import_cirq
from .gates import GateGenerator, gate_db
from .verification.command_registry import CommandRegistry

import copy
import json
from typing import Callable, Dict, Optional, List

from cirq import GateOperation
from qm.qua import switch_, case_, declare, align, for_
from qualang_tools.bakery.bakery import Baking, baking
from tqdm import tqdm

cirq = import_cirq()


class RBBaker:
    def __init__(
        self,
        config,
        single_qubit_gate_generator: Callable,
        two_qubit_gate_generators: Dict[str, Callable],
        interleaving_gate: Optional[List[cirq.GateOperation]] = None,
        command_registry: Optional[CommandRegistry] = None,
    ):
        self._command_registry = command_registry
        self._config = copy.deepcopy(config)
        self._single_qubit_gate_generator = single_qubit_gate_generator
        self._two_qubit_gate_generators = two_qubit_gate_generators
        self._interleaving_gate = interleaving_gate
        self._symplectic_generator = GateGenerator(set(two_qubit_gate_generators.keys()))
        self._all_elements = self._collect_all_elements()
        self._cmd_to_op = {}
        self._op_to_baking = {}

    @property
    def all_elements(self):
        return self._all_elements

    @staticmethod
    def _get_qubits(op):
        return [q.x for q in op.qubits]

    @staticmethod
    def _get_phased_xz_args(op):
        return op.gate.x_exponent, op.gate.z_exponent, op.gate.axis_phase_exponent

    def _validate_two_qubit_gate_available(self, name):
        if name not in self._two_qubit_gate_generators:
            raise RuntimeError(f"Two qubit gate '{name}' implementation not provided.")

    def _gen_gate(self, baker: Baking, gate_op: GateOperation):
        if isinstance(gate_op.gate, cirq.PhasedXZGate):
            self._single_qubit_gate_generator(baker, self._get_qubits(gate_op)[0], *self._get_phased_xz_args(gate_op))
        elif isinstance(gate_op.gate, cirq.ISwapPowGate) and gate_op.gate.exponent == 0.5:
            self._validate_two_qubit_gate_available("sqr_iSWAP")
            self._two_qubit_gate_generators["sqr_iSWAP"](baker, *self._get_qubits(gate_op))
        elif isinstance(gate_op.gate, cirq.CNotPowGate) and gate_op.gate.exponent == 1:
            self._validate_two_qubit_gate_available("CNOT")
            self._two_qubit_gate_generators["CNOT"](baker, *self._get_qubits(gate_op))
        elif isinstance(gate_op.gate, cirq.CZPowGate) and gate_op.gate.exponent == 1:
            self._validate_two_qubit_gate_available("CZ")
            self._two_qubit_gate_generators["CZ"](baker, *self._get_qubits(gate_op))
        else:
            raise RuntimeError("unsupported gate")

    def _collect_all_elements(self):
        config = copy.deepcopy(self._config)
        qes = set()
        for cmd_id, command in enumerate(gate_db.commands):
            if self._command_registry is not None:
                self._command_registry.set_current_command_id(cmd_id)
            with baking(config) as b:
                self._update_baking_from_cmd_id(b, cmd_id)
                qes.update(b.get_qe_set())
                b.update_config = False
        if self._interleaving_gate is not None:
            if self._command_registry is not None:
                self._command_registry.set_current_command_id(cmd_id + 1)
            with baking(config) as b:
                self._update_baking_from_gates(b, self._interleaving_gate)
                qes.update(b.get_qe_set())
                b.update_config = False
        if self._command_registry is not None:
            self._command_registry.finish()
        return qes

    def _update_baking_from_gates(self, b: Baking, gate_ops, elements=None):
        prev_gate_qubits = []
        for gate_op in gate_ops:
            gate_qubits = self._get_qubits(gate_op)
            if len(gate_qubits) != len(prev_gate_qubits) and elements is not None:
                b.align(*elements)
            prev_gate_qubits = gate_qubits
            self._gen_gate(b, gate_op)
        if elements is not None:
            b.align(*elements)

    def gates_from_cmd_id(self, cmd_id):
        if 0 <= cmd_id < len(gate_db.commands):
            gate_ops = self._symplectic_generator.generate(cmd_id)
        elif self._interleaving_gate is not None and cmd_id == len(gate_db.commands):  # Interleaving gate
            gate_ops = self._interleaving_gate
        else:
            raise RuntimeError("command out of range")
        return gate_ops

    def _update_baking_from_cmd_id(self, b: Baking, cmd_id, elements=None):
        gate_ops = self.gates_from_cmd_id(cmd_id)
        return self._update_baking_from_gates(b, gate_ops, elements)

    @staticmethod
    def _unique_baker_identifier_for_qe(b: Baking, qe: str):
        identifier = {"samples": b._samples_dict[qe], "info": b._qe_dict[qe]}
        return json.dumps(identifier)

    def _bake_all_ops(self, config: dict):
        waveform_id_per_qe = {qe: 0 for qe in self._all_elements}
        waveform_to_baking = {qe: {} for qe in self._all_elements}
        cmd_to_op = {qe: {} for qe in self._all_elements}
        op_to_baking = {qe: [] for qe in self._all_elements}
        num_of_commands = len(gate_db.commands) + (0 if self._interleaving_gate is None else 1)
        for cmd_id in tqdm(range(num_of_commands), desc="Baking pulses which comprise Cliffords", unit="command"):
            with baking(config) as b:
                self._update_baking_from_cmd_id(b, cmd_id, self._all_elements)
                any_qe_used = False
                for qe in self._all_elements:
                    key = self._unique_baker_identifier_for_qe(b, qe)
                    if key not in waveform_to_baking[qe]:
                        waveform_to_baking[qe][key] = waveform_id_per_qe[qe], b
                        op_to_baking[qe].append(b)
                        waveform_id_per_qe[qe] += 1
                        any_qe_used = True
                    cmd_to_op[qe][cmd_id] = waveform_to_baking[qe][key][0]
                b.update_config = any_qe_used
        return cmd_to_op, op_to_baking

    def bake(self) -> dict:
        config = copy.deepcopy(self._config)
        self._cmd_to_op, self._op_to_baking = self._bake_all_ops(config)
        return config

    def decode(self, cmd_id, element):
        return self._cmd_to_op[element][cmd_id]

    @staticmethod
    def _run_baking_for_qe(b: Baking, qe: str):
        orig_get_qe_set = b.get_qe_set
        b.get_qe_set = lambda: {qe}
        b.run()
        b.get_qe_set = orig_get_qe_set

    def run(self, op_list_per_qe: dict, length, unsafe=True):
        if set(op_list_per_qe.keys()) != self._all_elements:
            raise RuntimeError(f"must specify ops for all elements: {', '.join(self._all_elements)} ")

        align()
        for qe, op_list in op_list_per_qe.items():
            cmd_i = declare(int)
            with for_(cmd_i, 0, cmd_i < length, cmd_i + 1):
                with switch_(op_list[cmd_i], unsafe=unsafe):
                    for op_id, b in enumerate(self._op_to_baking[qe]):
                        with case_(op_id):
                            self._run_baking_for_qe(b, qe)
        align()



# flake8: noqa

import matplotlib.pyplot as plt
from qm.qua import *
from qm import QuantumMachinesManager
from qualang_tools.bakery.bakery import Baking

from qualang_tools.characterization.two_qubit_rb import TwoQubitRb, TwoQubitRbDebugger


##############################
## General helper functions ##
##############################
def multiplexed_readout(I, I_st, Q, Q_st, resonators, sequential=False, amplitude=1.0, weights=""):
    """Perform multiplexed readout on two resonators"""
    if type(resonators) is not list:
        resonators = [resonators]

    for ind, res in enumerate(resonators):
        measure(
            "readout" * amp(amplitude),
            f"rr{res}",
            None,
            dual_demod.full(weights + "cos", weights + "sin", I[ind]),
            dual_demod.full(weights + "minus_sin", weights + "cos", Q[ind]),
        )

        if I_st is not None:
            save(I[ind], I_st[ind])
        if Q_st is not None:
            save(Q[ind], Q_st[ind])

        if sequential and ind < len(resonators) - 1:
            align(f"rr{res}", f"rr{res+1}")


##############################
##  Two-qubit RB functions  ##
##############################
# assign a string to a variable to be able to call them in the functions
q1_idx_str = "1"
q2_idx_str = "2"


# single qubit generic gate constructor Z^{z}Z^{a}X^{x}Z^{-a}
# that can reach any point on the Bloch sphere (starting from arbitrary points)
def bake_phased_xz(baker: Baking, q, x, z, a):
    if q == 1:
        element = f"q{q1_idx_str}_xy"
    else:
        element = f"q{q2_idx_str}_xy"

    baker.frame_rotation_2pi(a / 2, element)
    baker.play("x180", element, amp=x)
    baker.frame_rotation_2pi(-(a + z) / 2, element)


# single qubit phase corrections in units of 2pi applied after the CZ gate
qubit1_frame_update = 0.23  # example values, should be taken from QPU parameters
qubit2_frame_update = 0.12  # example values, should be taken from QPU parameters


# defines the CZ gate that realizes the mapping |00> -> |00>, |01> -> |01>, |10> -> |10>, |11> -> -|11>
def bake_cz(baker: Baking, q1, q2):
    q1_xy_element = f"q{q1_idx_str}_xy"
    q2_xy_element = f"q{q2_idx_str}_xy"
    q1_z_element = f"q{q1_idx_str}_z"

    baker.play("cz", q1_z_element)
    baker.align()
    baker.frame_rotation_2pi(qubit1_frame_update, q1_xy_element)
    baker.frame_rotation_2pi(qubit2_frame_update, q2_xy_element)
    baker.align()


def prep():
    wait(int(10 * qubit_T1))  # thermal preparation in clock cycles (time = 10 x T1 x 4ns)
    align()


def meas():
    threshold1 = 0.3  # threshold for state discrimination 0 <-> 1 using the I quadrature
    threshold2 = 0.3  # threshold for state discrimination 0 <-> 1 using the I quadrature
    I1 = declare(fixed)
    I2 = declare(fixed)
    Q1 = declare(fixed)
    Q2 = declare(fixed)
    state1 = declare(bool)
    state2 = declare(bool)
    multiplexed_readout(
        [I1, I2], None, [Q1, Q2], None, resonators=[1, 2], weights="rotated_"
    )  # readout macro for multiplexed readout
    assign(state1, I1 > threshold1)  # assume that all information is in I
    assign(state2, I2 > threshold2)  # assume that all information is in I
    return state1, state2


##############################
##  Two-qubit RB execution  ##
##############################
qmm = QuantumMachinesManager(host="enter-your-qmm-ip-address", port=None, cluster_name="enter-your-cluster-name")

# create RB experiment from configuration and defined functions
rb = TwoQubitRb(
    config=config,  # enter your QUA config here
    single_qubit_gate_generator=bake_phased_xz,
    two_qubit_gate_generators={"CZ": bake_cz},  # can also provide e.g. "CNOT": bake_cnot
    prep_func=prep,
    measure_func=meas,
    interleaving_gate=None,
    # interleaving_gate=[cirq.CZ(cirq.LineQubit(0), cirq.LineQubit(1))],
    verify_generation=False,
)

# run simpler experiment to verify `bake_phased_xz`, `prep` and `meas`
# rb_debugger = TwoQubitRbDebugger(rb)
# rb_debugger.run_phased_xz_commands(qmm, 20)
# plt.show()

# run 2Q-RB experiment
res = rb.run(
    qmm,
    circuit_depths=[1, 2, 3, 4, 5],
    num_circuits_per_depth=2,
    num_shots_per_circuit=1,
    # unsafe=True will minimize switch-case gaps, but can lead to unexpected behaviour
    unsafe=False,
)

# circuit_depths ~ how many consecutive Clifford gates within one executed circuit
# (https://qiskit.org/documentation/apidoc/circuit.html)
# num_circuits_per_depth ~ how many random circuits within one depth
# num_shots_per_circuit ~ repetitions of the same circuit (averaging)

res.plot_hist()
plt.show()

res.plot_fidelity()
plt.show()

# verify/save the random sequences created during the experiment
rb.save_sequences_to_file("sequences.txt")  # saves the gates used in each random sequence
rb.save_command_mapping_to_file("commands.txt")  # saves mapping from "command id" to sequence
# rb.print_sequence()
# rb.print_command_mapping()
# rb.verify_sequences()  # simulates random sequences to ensure they recover to ground state. takes a while...

# # get the interleaved gate fidelity
# from two_qubit_rb.RBResult import get_interleaved_gate_fidelity
# interleaved_gate_fidelity = get_interleaved_gate_fidelity(
#     num_qubits=2,
#     reference_alpha=0.12345,  # replace with value from prior, non-interleaved experiment
#     # interleaved_alpha=res.fit_exponential()[1],  # alpha from the interleaved experiment
# )
# print(f"Interleaved Gate Fidelity: {interleaved_gate_fidelity*100:.3f}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import warnings

from dash import dcc
import dash_bootstrap_components as dbc
from dash import html
from dash import Input, Output
from .server.app import app
from .server.editors import *
from .server.upload import *
from .server.download import *
from .server import config_editor as ced
import webbrowser
from threading import Timer
import argparse


from waitress import serve


@app.callback(Output("page-content", "children"), [Input("url", "pathname")])
def render_page_content(pathname):
    pathname = str(pathname)

    if pathname == "/":
        return upload
    elif pathname.startswith("/config"):
        return config_editor(pathname)
    elif pathname == "/download":
        return download_page
    # If the user tries to reach a different page, return a 404 message
    return html.H1("404: Not found", className="text-danger")


SIDEBAR_STYLE = {
    "position": "fixed",
    "top": 0,
    "left": 0,
    "bottom": 0,
    "width": "16rem",
    "padding": "2rem 1rem",
    "background-color": "#f8f9fa",
}

sidebar = html.Div(
    [
        html.Center(html.H3("ConfigBuilder")),
        html.Hr(),
        dbc.Nav(
            [
                dbc.NavLink(
                    [html.I(className="bi bi-upload me-2"), "Upload or Start new"],
                    href="/",
                    active="exact",
                ),
                dbc.NavLink(
                    [html.I(className="bi bi-sliders me-2"), "View & Edit "],
                    href="/config",
                    active="exact",
                ),
                dbc.NavLink(
                    [html.I(className="bi bi-download me-2"), "Download final"],
                    href="/download",
                    active="exact",
                ),
            ],
            vertical=True,
            pills=True,
        ),
    ],
    style=SIDEBAR_STYLE,
)

CONTENT_STYLE = {
    "margin-left": "18rem",
    "margin-right": "2rem",
    "padding": "2rem 1rem",
}

content = html.Div(id="page-content", style=CONTENT_STYLE)

app.layout = html.Div([dcc.Location(id="url"), sidebar, content])
guiserver = app.server


def open_browser():
    global port_number
    webbrowser.open_new(f"http://localhost:{port_number}")


if __name__ == "__main__":
    warnings.warn(
        "ConfigGUI is no longer being actively developed and may not support all config functionality",
        DeprecationWarning,
        stacklevel=2,
    )
    global port_number

    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--port", help="port number to use for GUI server", type=int)
    args = parser.parse_args()
    if args.port is not None:
        port_number = args.port
    else:
        port_number = 8051
    host = "127.0.0.1"

    timer = Timer(3, open_browser)
    timer.start()
    try:
        print(f"\tConfig GUI is running on http://{host}:{port_number}")
        print("\tIf web browser does not open, please enter the address stated above manually.")
        print("\tStop GUI by pressing Ctrl+C")
        print(
            "\t%s: %s"
            % (
                ced.config_structure["info"]["title"],
                ced.config_structure["info"]["version"],
            )
        )
        serve(app.server, host=host, port=port_number)
        # for development:
        # app.run_server(host="127.0.0.1", debug=False, port=port_number)
    except OSError as e:
        timer.cancel()
        if e.args[0] != 98:
            # if the error is not "address already in use" (trying to bind
            # to a port on which another application is already listening)
            # pass directly the error to the user
            raise e
        print("Port number 8051 is already in use")
        print("Please specify alternative port number as follows:")
        print("python -m qualang_tools.config.gui -p PORT_NUMBER")



"""calling function libraries"""

import copy
import math
from time import sleep

import numpy as np
from qm import QuantumMachine
from qm import QmJob
from qm import QuantumMachinesManager
from qm.qua import *


def _round_to_fixed_point_accuracy(x, accuracy=2**-16):
    return round(x / accuracy) * accuracy


def _floor_to_fixed_point_accuracy(x, accuracy=2**-16):
    return math.floor(x / accuracy) * accuracy


class ManualOutputControl:
    """
    Creates a control panel to turn on and off the analog and digital channels of the OPX.
    """

    def __init__(
        self,
        configuration,
        host=None,
        port=None,
        close_previous=True,
        elements_to_control=None,
    ):
        """
        Gets a QUA configuration file and creates different Quantum Machines. One QM continuously runs all the
        analog ports and the rest continuously runs all the digital ports. This enables controlling the amplitude and
        frequency of the analog ports and turning on and off the digital ports.
        All digital outputs start turned off, and all analog outputs start with zero amplitude, but with the frequency
        from the configuration.

        :param configuration: A configuration file containing the elements to control.
        :param str host: The host or IP of the QOP. Defaults to `None`: local settings are used.
        :param str port: The port used to access the QOP. Defaults to `None`, local settings are used.
        :param bool close_previous: Close currently running Quantum Machines. Note that if False, and a Quantum Machine
                                    which uses the same ports is already open, then this function would fail.
        :param array-like elements_to_control: A list of elements to be controlled.
                                    If empty, all elements in the config are included.
        """
        self.qmm = QuantumMachinesManager(host=host, port=port)
        if close_previous:
            self.qmm.close_all_quantum_machines()
        self.analog_config = None
        self.digital_configs = None
        self.analog_elements = []
        self.digital_elements = []
        self.digital_index = {}
        self.analog_data = {}
        self.digital_data = {}
        self.analog_job = None
        self.digital_jobs = None
        self.ANALOG_WAVEFORM_AMPLITUDE = 0.5 - 2**-16
        self._process_config(configuration, elements_to_control)
        self.analog_qm = self.qmm.open_qm(self.analog_config, False)
        self._start_digital_qms()
        self._start_analog()

    @classmethod
    def ports(
        cls,
        analog_ports=None,
        digital_ports=None,
        host=None,
        port=None,
        close_previous=True,
    ):
        """
        Gets a list of analog and digital channels and creates different Quantum Machines. One QM continuously runs all
        the analog ports and the rest continuously runs all the digital ports. This enables controlling the amplitude
        and frequency of the analog ports and turning on and off the digital ports.
        All digital outputs start turned off, and all analog outputs start with zero amplitude and zero frequency.

        :param list analog_ports: The list of analog ports to control. A tuple creates an IQ pair.
            For example, [1, 2, (3, 4)] creates two independent channels at ports 1 & 2, and one IQ pair at ports 3 & 4.
            For multiple controllers, increase the port by 10. For example, controller 3 port 2 should be 32.
        :param list digital_ports: The list of digital ports to control.
            For multiple controllers, increase the port by 10. For example, controller 3 port 2 should be 32.
        :param str host: The host or IP of the QOP. Defaults to `None`: local settings are used.
        :param str port: The port used to access the QOP. Defaults to `None`, local settings are used.
        :param bool close_previous: Close currently running Quantum Machines. Note that if False, and a Quantum Machine
                                    which uses the same ports is already open, then this function would fail.
        """
        if analog_ports is None and digital_ports is None:
            raise Exception("No ports specified")
        config = {"version": 1, "controllers": {}, "elements": {}}
        if analog_ports is not None:
            for port_int in analog_ports:
                if isinstance(port_int, tuple):
                    if len(port_int) == 2:
                        con_number = (port_int[0] - 1) // 10 + 1
                        if con_number is not (port_int[1] - 1) // 10 + 1:
                            raise Exception(f"Ports {port_int[0]} and {port_int[1]} are not from the same controller")
                    else:
                        raise Exception(f"Port {port_int} should be either an integer or a tuple of two integers")
                else:
                    con_number = (port_int - 1) // 10 + 1
                con = f"con{con_number}"
                if con not in config["controllers"]:
                    config["controllers"][con] = {
                        "type": "opx1",
                        "analog_outputs": {},
                        "digital_outputs": {},
                    }

                if isinstance(port_int, tuple):
                    port_str = str(port_int)
                    port1 = (port_int[0] - 1) % 10 + 1
                    port2 = (port_int[1] - 1) % 10 + 1
                    if port1 not in config["controllers"][con]["analog_outputs"]:
                        config["controllers"][con]["analog_outputs"][port1] = {"offset": 0.0}
                    if port2 not in config["controllers"][con]["analog_outputs"]:
                        config["controllers"][con]["analog_outputs"][port2] = {"offset": 0.0}
                    if port_str not in config["elements"]:
                        config["elements"][port_str] = {
                            "mixInputs": {
                                "I": (con, port1),
                                "Q": (con, port2),
                            },
                            "intermediate_frequency": 0e6,
                        }
                else:
                    port_str = str(port_int)
                    port_int = (port_int - 1) % 10 + 1
                    if port_int not in config["controllers"][con]["analog_outputs"]:
                        config["controllers"][con]["analog_outputs"][port_int] = {"offset": 0.0}
                    if port_str not in config["elements"]:
                        config["elements"][port_str] = {
                            "singleInput": {
                                "port": (con, port_int),
                            },
                            "intermediate_frequency": 0e6,
                        }

        if digital_ports is not None:
            for port_int in digital_ports:
                con_number = (port_int - 1) // 10 + 1
                port_str = str(port_int)
                port_int = (port_int - 1) % 10 + 1
                con = f"con{con_number}"
                if con not in config["controllers"]:
                    config["controllers"][con] = {
                        "type": "opx1",
                        "analog_outputs": {},
                        "digital_outputs": {},
                    }
                if port_int not in config["controllers"][con]["digital_outputs"]:
                    config["controllers"][con]["digital_outputs"][port_int] = {}
                if port_str not in config["elements"]:
                    config["elements"][port_str] = {
                        "digitalInputs": {
                            "input": {
                                "port": (con, port_int),
                                "delay": 0,
                                "buffer": 0,
                            }
                        }
                    }
                else:
                    config["elements"][port_str]["digitalInputs"] = {
                        "input": {
                            "port": (con, port_int),
                            "delay": 0,
                            "buffer": 0,
                        }
                    }

        return cls(
            config,
            host=host,
            port=port,
            close_previous=close_previous,
            elements_to_control=None,
        )

    def turn_on_element(self, element, amplitude=None, frequency=None):
        """
        Turns on the digital and analog output of a given element.
        If no amplitude is given, uses maximum amplitude.
        If no frequency is given, uses existing frequency.

        :param element: An element to be turned on.
        :param amplitude:The amplitude of the analog output of the element.
        :param frequency: The frequency of the analog output of the element.
        """
        if not isinstance(element, str):
            element = str(element)
        if element not in self.digital_elements and element not in self.analog_elements:
            raise Exception(f"Element {element} is not part of the elements in the configuration files")
        if amplitude is None:
            amplitude = self.ANALOG_WAVEFORM_AMPLITUDE
        if frequency is not None:
            self.set_frequency(element, frequency, ignore_missing_elements=True)
        self.set_amplitude(element, amplitude, ignore_missing_elements=True)
        self.digital_on(element, ignore_missing_elements=True)

    def turn_off_elements(self, *elements):
        """
        Turns off the digital and analog for all the given elements.
        If no elements are given, turns off all elements.

        :param elements: A variable number of elements to be turned off.
        """
        if not len(elements) == 0:
            for element in elements:
                if not isinstance(element, str):
                    element = str(element)
                if element not in self.digital_elements and element not in self.analog_elements:
                    raise Exception(f"Element {element} is not part of the elements in the configuration files")
        self.digital_off(*elements, ignore_missing_elements=True)
        if len(elements) == 0:
            elements = self.analog_elements
        for element in elements:
            self.set_amplitude(element, 0, ignore_missing_elements=True)

    def set_amplitude(self, element, value, ignore_missing_elements=False):
        """
        Sets the amplitude of an analog element.

        :param str element: The name of the analog element to be updated
        :param float value: The new amplitude of the analog element
        :param bool ignore_missing_elements: If false (default), throws an error if the element is not found.
            If false, simply returns the function.
        """
        if not isinstance(element, str):
            element = str(element)
        if element not in self.analog_elements:
            if ignore_missing_elements:
                return
            else:
                raise Exception(f"Element {element} is not part of the elements in the analog configuration file")
        if abs(value) > self.ANALOG_WAVEFORM_AMPLITUDE:
            if value == 0.5:
                value = 0.5 - 2**-16
            elif value == -0.5:
                value = -0.5 + 2**-16
            else:
                raise Exception(f"The absolute value of the amplitude must smaller than 0.5, {value} was given")

        prev_value = self.analog_data[element]["amplitude"]
        delta_value = (value - prev_value) * (1 / self.ANALOG_WAVEFORM_AMPLITUDE)
        delta_value = _round_to_fixed_point_accuracy(delta_value)
        if delta_value == 0:
            return
        self.analog_data[element]["amplitude"] = _floor_to_fixed_point_accuracy(
            prev_value + delta_value * self.ANALOG_WAVEFORM_AMPLITUDE
        )

        while not self.analog_job.is_paused():
            sleep(0.01)

        self.analog_qm.set_io_values(
            int(self.analog_elements.index(element)) + len(self.analog_elements),
            float(delta_value),
        )
        self.analog_job.resume()

    def set_frequency(self, element, value, ignore_missing_elements=False):
        """
        Sets the frequency of an analog element.

        :param str element: the name of the analog element to be updated
        :param int value: the new frequency of the analog element
        :param bool ignore_missing_elements: If false (default), throws an error if the element is not found.
            If false, simply returns the function.
        """
        if not isinstance(element, str):
            element = str(element)
        if element not in self.analog_elements:
            if ignore_missing_elements:
                return
            else:
                raise Exception(f"Element {element} is not part of the elements in the analog configuration file")
        if int(value) > int(500e6):
            raise Exception(f"The frequency should be lower than 500e6, {value} was given")
        self.analog_data[element]["frequency"] = value
        while not self.analog_job.is_paused():
            sleep(0.01)
        self.analog_qm.set_io_values(int(self.analog_elements.index(element)), int(value))
        self.analog_job.resume()

    def digital_on(self, *digital_elements, ignore_missing_elements=False):
        """
        Turns on all digital elements inputted by the user.
        If no input is given, turns on all elements.

        :param digital_elements: A variable number of elements to be turned on.
        :param bool ignore_missing_elements: If false (default), throws an error if the element is not found.
            If false, simply returns the function.
        """
        if len(digital_elements) == 0:
            digital_elements = self.digital_elements
        for element in digital_elements:
            if not isinstance(element, str):
                element = str(element)
            if element not in self.digital_elements:
                if ignore_missing_elements:
                    return
                else:
                    raise Exception(f"Element {element} is not part of the elements in the digital configuration file")
            self.digital_data[element] = True
        self._start_digital()

    def digital_off(self, *digital_elements, ignore_missing_elements=False):
        """
        Turns off all digital elements inputted by the user.
        If no input is given, turns off all elements.

        :param digital_elements: A variable number of elements to be turned off.
        :param bool ignore_missing_elements: If false (default), throws an error if the element is not found.
            If false, simply returns the function.
        """
        if len(digital_elements) == 0:
            digital_elements = self.digital_elements
        for element in digital_elements:
            if not isinstance(element, str):
                element = str(element)
            if element not in self.digital_elements:
                if ignore_missing_elements:
                    return
                else:
                    raise Exception(f"Element {element} is not part of the elements in the digital configuration file")
            self.digital_data[element] = False
        self._start_digital()

    def digital_switch(self, *digital_element):
        """
        Switches the state of the given digital elements from on to off, and from off to on.

        :param digital_element: A variable number of elements to be switched.
        """
        if len(digital_element) == 0:
            return
        for element in digital_element:
            if not isinstance(element, str):
                element = str(element)
            if element not in self.digital_elements:
                raise Exception(f"Element {element} is not part of the elements in the digital configuration file")
            self.digital_data[element] = not self.digital_data[element]
        self._start_digital()

    def analog_status(self):
        """
        Returns a dictionary of the analog elements, with their current amplitude and frequency.
        """
        return self.analog_data

    def digital_status(self):
        """
        Returns a dictionary of the digital elements, with their current status.
        """
        return self.digital_data

    def print_analog_status(self):
        """
        Prints a list of the analog elements, with their current amplitude and frequency.
        """
        for element in self.analog_elements:
            print(
                element
                + " - Amplitude: "
                + str(self.analog_data[element]["amplitude"])
                + ", Frequency: "
                + str(self.analog_data[element]["frequency"])
            )

    def print_digital_status(self):
        """
        Prints a list of the digital elements, with a True (False) to indicate the element is on (off).
        """
        for element in self.digital_elements:
            print(element + " - " + str(self.digital_data[element]))

    def close(self):
        """
        Halts all jobs sent to the OPX and then closes the quantum machine.
        """
        for con_jobs in self.digital_jobs:
            for job in con_jobs:
                if job is not None:
                    job.halt()
        for con_qms in self.digital_qms:
            for qm in con_qms:
                if qm is not None:
                    qm.close()
        self.analog_job.halt()
        self.analog_qm.close()
        self.qmm.close()

    def _process_config(self, config_original, elements_to_control=None):
        """
        This function creates multiple separate configuration files, on containing only analog elements, and the others
        only contains digital elements.

        :param config_original: a QUA configuration dictionary which contain all defined analog and digital elements.
        :param elements_to_control: A list of elements to be controlled.
                                    If empty, all elements in the config are included.
        """
        config = copy.deepcopy(config_original)
        self.analog_config = {}
        digital_config = {}
        pulser_count = {}

        highest_con = 0
        for con in list(config["controllers"].keys()):
            highest_con = int(con[-1]) if int(con[-1]) > highest_con else highest_con
        self.digital_configs = np.empty((highest_con, 10), dtype=dict)
        self.digital_jobs = np.empty((highest_con, 10), dtype=QmJob)
        self.digital_qms = np.empty((highest_con, 10), dtype=QuantumMachine)

        # Analog config
        self.analog_config["version"] = 1
        self.analog_config["controllers"] = {}
        for controller in list(config["controllers"].keys()):
            self.analog_config["controllers"][controller] = {
                "type": "opx1",
                "analog_outputs": {},
            }
            self.analog_config["controllers"][controller]["analog_outputs"] = config["controllers"][controller][
                "analog_outputs"
            ]
            pulser_count[controller] = 0
        self.analog_config["elements"] = {}
        self.analog_config["waveforms"] = {
            "zero_wf": {"type": "constant", "sample": 0.0},
            "const_wf": {"type": "constant", "sample": self.ANALOG_WAVEFORM_AMPLITUDE},
        }
        self.analog_config["pulses"] = {}
        self.analog_config["pulses"]["single_on"] = {
            "operation": "control",
            "length": 16,
            "waveforms": {"single": "const_wf"},
        }
        self.analog_config["pulses"]["IQ_Ion"] = {
            "operation": "control",
            "length": 16,
            "waveforms": {"I": "const_wf", "Q": "zero_wf"},
        }
        if config.get("mixers") is not None:
            self.analog_config["mixers"] = config["mixers"]

        # Digital config
        digital_config["version"] = 1
        digital_config["controllers"] = {}
        digital_config["elements"] = {}
        digital_config["pulses"] = {
            "digital_ON": {
                "digital_marker": "ON",
                "length": 1000,
                "operation": "control",
            }
        }
        digital_config["digital_waveforms"] = {"ON": {"samples": [(1, 0)]}}

        # Elements
        if elements_to_control is None:
            elements = list(config["elements"].keys())
        else:
            elements = elements_to_control
        # Digital
        for element in elements:
            if config["elements"][element].get("digitalInputs") is not None:
                self.digital_index[element] = []
                self.digital_data[element] = False
                self.digital_elements.append(element)
                for digital_input in config["elements"][element]["digitalInputs"]:
                    con, port = config["elements"][element]["digitalInputs"][digital_input]["port"]
                    con_int_index = int(con[-1]) - 1
                    port_index = port - 1
                    self.digital_index[element].append((con_int_index, port_index))
                    if self.digital_configs[con_int_index][port_index] is None:
                        pulser_count[con] += 1
                        self.digital_configs[con_int_index][port_index] = copy.deepcopy(digital_config)
                        self.digital_configs[con_int_index][port_index]["controllers"][con] = {
                            "type": "opx1",
                            "digital_outputs": {port: {}},
                        }
                        self.digital_configs[con_int_index][port_index]["elements"]["digital"] = {
                            "operations": {"ON": "digital_ON"}
                        }
                        self.digital_configs[con_int_index][port_index]["elements"]["digital"]["digitalInputs"] = {
                            "digital": {
                                "port": (con, port),
                                "delay": 0,
                                "buffer": 0,
                            },
                        }
        # Analog
        for element in elements:
            if config["elements"][element].get("mixInputs") is not None:
                con = config["elements"][element]["mixInputs"]["I"][0]
                pulser_count[con] += 2
                self.analog_config["elements"][element] = config["elements"][element]
                if self.analog_config["elements"][element].get("digitalInputs") is not None:
                    self.analog_config["elements"][element].pop("digitalInputs")
                if self.analog_config["elements"][element].get("outputs") is not None:
                    self.analog_config["elements"][element].pop("outputs")
                    self.analog_config["elements"][element].pop("time_of_flight")
                    self.analog_config["elements"][element].pop("smearing")
                if self.analog_config["elements"][element].get("operations") is not None:
                    self.analog_config["elements"][element].pop("operations")
                self.analog_config["elements"][element]["operations"] = {"play": "IQ_Ion"}
                self.analog_elements.append(element)

            elif config["elements"][element].get("singleInput") is not None:
                con = config["elements"][element]["singleInput"]["port"][0]
                pulser_count[con] += 1
                self.analog_config["elements"][element] = config["elements"][element]
                if self.analog_config["elements"][element].get("digitalInputs") is not None:
                    self.analog_config["elements"][element].pop("digitalInputs")
                if self.analog_config["elements"][element].get("outputs") is not None:
                    self.analog_config["elements"][element].pop("outputs")
                    self.analog_config["elements"][element].pop("time_of_flight")
                    self.analog_config["elements"][element].pop("smearing")
                if self.analog_config["elements"][element].get("operations") is not None:
                    self.analog_config["elements"][element].pop("operations")
                self.analog_config["elements"][element]["operations"] = {"play": "single_on"}
                self.analog_elements.append(element)
        for element in self.analog_elements:
            self.analog_config["elements"][element]["hold_offset"] = {"duration": 16}
            self.analog_data[element] = {
                "amplitude": 0,
                "frequency": self.analog_config["elements"][element].get("intermediate_frequency"),
            }

        opx_plus = True if self.qmm.version()["server"][0] == "2" else False
        for con in list(config["controllers"].keys()):
            if opx_plus and pulser_count[con] > 18:
                raise Exception(
                    f"Given configuration requires {pulser_count[con]} threads in {con}, but only {18} are available"
                )
            elif not opx_plus and pulser_count[con] > 10:
                raise Exception(
                    f"Given configuration requires {pulser_count[con]} threads in {con}, but only {10} are available"
                )

    def _start_analog(self):
        """
        Creates and starts QUA program that initialize the analog elements at 0 amplitude and awaits IO variables to
        update the amplitude and frequency of each element
        """

        with program() as prog:
            io_var1 = declare(int)
            for element in self.analog_elements:
                play("play" * amp(0), element)
            with infinite_loop_():
                pause()
                assign(io_var1, IO1)
                self._QUA_update_freq_or_amp(io_var1, IO2)

        self.analog_job = self.analog_qm.execute(prog)

    def _start_digital_qms(self):
        """
        Opens up multiple Quantum Machines to control the digital outputs
        """
        for con_port_list in self.digital_index.values():
            for con, port in con_port_list:
                if self.digital_qms[con][port] is None:
                    self.digital_qms[con][port] = self.qmm.open_qm(self.digital_configs[con][port], False)

    def _start_digital(self):
        """
        Creates and starts QUA program that is used to run the digital elements in an infinite loop.
        """
        with program() as prog:
            save(1, "a")
            with infinite_loop_():
                play("ON", "digital")
        digital_on = []

        # Create a list of ports that needs to be on
        for element in self.digital_elements:
            for con, port in self.digital_index[element]:
                if self.digital_data[element]:  # Needs to be on
                    digital_on.append((con, port))

        for con_port_list in self.digital_index.values():
            for con, port in con_port_list:
                if (con, port) in digital_on:  # Needs to be on
                    if (  # Is not on
                        self.digital_jobs[con][port] is None
                        or not self.digital_jobs[con][port].result_handles.is_processing()
                    ):
                        self.digital_jobs[con][port] = self.digital_qms[con][port].execute(prog)
                    else:  # Is already on
                        pass
                else:  # Needs to be off
                    if self.digital_jobs[con][port] is not None:
                        self.digital_jobs[con][port].halt()

    def _QUA_update_freq_or_amp(self, input1, input2):
        """
        This function updates the amplitude or frequency of an element

        :param QUA Int input1: Indicates which element to update and whether it is amplitude or frequency using the
        following method: If there are 5 elements: ['a', 'b', 'c', 'd', 'e'], and input1 is '0', then it would change
            the amplitude of the first ('a'). If it input1 is '6', it would change the frequency of the 2nd ('b').
        :param QUA IO input2: The new frequency or delta amplitude to update them element.
        """
        with if_(input1 >= len(self.analog_elements)):
            assign(input1, input1 - len(self.analog_elements))
            a = declare(fixed)
            assign(a, input2)
            with switch_(input1):
                for i in range(len(self.analog_elements)):
                    with case_(i):
                        play("play" * amp(a), self.analog_elements[i])
        with else_():
            freq = declare(int)
            assign(freq, input2)
            with switch_(input1):
                for i in range(len(self.analog_elements)):
                    with case_(i):
                        if (
                            self.analog_config["elements"][self.analog_elements[i]].get("intermediate_frequency")
                            is not None
                        ):
                            update_frequency(self.analog_elements[i], freq)



from pathlib import Path
from typing import Union, Callable, Literal

from qualang_tools.bakery.bakery import Baking

from .gates import PhasedXZ, CZ, CNOT, Gate

Command = list[Gate]


class CommandRegistry:
    """
    Dataclass to track which single- or two-qubit gates are being baked in order
    to construct a two-qubit randomized benchmarking command.
    """

    def __init__(self):
        self._current_command_id = 0
        self._commands: dict[int, Command] = {}
        self._is_finished = False

    def register_phase_xz(self, q, x, z, a):
        gate = PhasedXZ(q=q, z=z, x=x, a=a)
        self._register_gate(gate)

    def register_cz(self):
        gate = CZ()
        self._register_gate(gate)

    def register_cnot(self, q):
        gate = CNOT(q=q)
        self._register_gate(gate)

    def _register_gate(self, gate: Gate):
        if self.is_finished():
            return
        if self._current_command_id in self._commands:
            command_list = self._commands[self._current_command_id]
        else:
            command_list = []
            self._commands[self._current_command_id] = command_list
        command_list.append(gate)

    def get_command_by_id(self, command_id: int):
        return self._commands[command_id]

    def set_current_command_id(self, command_id: int):
        self._current_command_id = command_id

    def _serialize_commands(self):
        result = ""
        for i in self._commands.keys():
            result += f"Command {i}:\n"
            for j, gate in enumerate(self._commands[i]):
                result += f"\t{j}: {gate}\n"
            result += "\n"
        return result

    def print_commands(self):
        print(self._serialize_commands())

    def save_to_file(self, path: Union[str, Path]):
        with open(path, "w+") as f:
            f.write(self._serialize_commands())

    def finish(self):
        """disable the incidental recording of any more commands."""
        self._is_finished = True

    def is_finished(self):
        return self._is_finished


PhasedXZGeneratorFunc = Callable[[Baking, int, float, float, float], None]
SingleQubitGateGeneratorFunc = Union[PhasedXZGeneratorFunc]


def decorate_single_qubit_generator_with_command_recording(
    single_qubit_gate_generator: PhasedXZGeneratorFunc, command_registry: CommandRegistry
) -> PhasedXZGeneratorFunc:
    """
    Decorates the `single_qubit_gate_generator` function so that it records
    every function call an input parameters it receives and registers them
    with the provided `command_registry`.
    """

    def decorated_generator(baker: Baking, q: int, x: float, z: float, a: float):
        command_registry.register_phase_xz(q=q, x=x, z=z, a=a)
        return single_qubit_gate_generator(baker, q, x, z, a)

    return decorated_generator


CZGeneratorFunc = Callable[[Baking, int, int], None]
CNOTGeneratorFunc = Callable[[Baking, int, int], None]
TwoQubitGateGeneratorFunc = Union[CZGeneratorFunc, CNOTGeneratorFunc]
TwoQubitGateGenerator = dict[Literal["CZ", "CNOT"], TwoQubitGateGeneratorFunc]


def decorate_two_qubit_gate_generator_with_command_recording(
    two_qubit_gate_generator: TwoQubitGateGenerator, command_registry: CommandRegistry
) -> TwoQubitGateGenerator:
    """
    Decorates the `single_qubit_gate_generator` function so that it
    records every call to itself and registers the corresponding input parameters.
    """
    decorated_two_qubit_gate_generator = {}
    for gate_name, gate_generator in two_qubit_gate_generator.items():
        if gate_name == "CZ":

            def decorated_generator(*args):
                command_registry.register_cz()
                return gate_generator(*args)

        elif gate_name == "CNOT":

            def decorated_generator(*args):
                control_qubit_index = args[1]
                command_registry.register_cnot(q=control_qubit_index)
                return gate_generator(*args)

        else:
            raise NotImplementedError(
                f"Command recording not implemented for {gate_name}. Please contact customer success."
            )

        decorated_two_qubit_gate_generator[gate_name] = decorated_generator

    return decorated_two_qubit_gate_generator



import abc
from typing import Union, List, Tuple, Optional, Dict

from qualang_tools.config.exceptions import ConfigurationError, _warn_if_not_none
from qualang_tools.config.primitive_components import (
    ConfigBuilderElement,
    AnalogInputPort,
    DigitalOutputPort,
    Waveform,
    DigitalSample,
    Pulse,
    Weights,
    AnalogOutputPort,
    DigitalInputPort,
    Operation,
    IntegrationWeights,
    MixerData,
)


class Controller(ConfigBuilderElement):
    def __init__(
        self,
        name: str,
        controller_type: str = "opx1",
    ):
        """A QOP controller

        :param name: Name for this controller
        :type name: str
        :param controller_type: defaults to "opx1"
        :type controller_type: str, optional
        """
        super(Controller, self).__init__(name)
        self.dict = dict()
        self.dict["type"] = controller_type
        self.controller_type = controller_type
        self.dict["analog_outputs"] = {}
        self.analog_output_ports = []
        self.dict["analog_inputs"] = {}
        self.analog_input_ports = []
        self.dict["digital_outputs"] = {}
        self.digital_output_ports = []
        self.dict["digital_inputs"] = {}
        self.digital_input_ports = []

    def analog_output(self, port: int, offset: float = 0):
        """Returns an instance of AnalogOutputPort associated  with a specific port number and offset if already in the configuration.
        otherwise, opens a new instance with the given port number.

        :param port: physical output port numebr
        :type port: int
        :param offset: sets the voltage offset of that port, defaults to 0
        :type offset: float, optional
        :return: An instance of the analog port
        :rtype: AnalogOutputPort
        """
        for i, p in enumerate(self.analog_output_ports):
            if port == p.info[1]:
                p.offset = offset
                return p
        self._use_analog_output_port(port, offset)
        return self.analog_output_ports[-1]

    def analog_input(self, port: int, offset: float = 0):
        """Returns an instance of AnalogInputPort with a specific port number and offset if already in the configuration.
        otherwise, opens a new instance with the given port number and offset.

        :param port: physical input port numebr
        :type port: int
        :param offset: [description], defaults to 0
        :type offset: float, optional
        :return: An instance of the input port
        :rtype: AnalogInputPort
        """
        for i, p in enumerate(self.analog_input_ports):
            if port == p.info[1]:
                p.offset = offset
                return p
        self._use_analog_input_port(port, offset)
        return self.analog_input_ports[-1]

    def digital_output(self, port: int):
        """Returns an instance of DigitalOutputPort with a specific port number and offset if already in the configuration.
        otherwise, opens a new instance with the given port number and offset.

        :param port: port number in the range 1-10
        :type port: int
        :return:
        :rtype: DigitalOutputPort
        """
        for i, p in enumerate(self.digital_output_ports):
            if port == p.info[1]:
                return p
        self._use_digital_output_port(port)
        return self.digital_output_ports[-1]

    def digital_input(self, port: int):
        """Returns an instance of DigitalInputPort with a specific port number if already in the configuration.
        otherwise, opens a new instance with the given port number.

        :param port: port number in the range 1-10
        :type port: int
        :return:
        :rtype: DigitalOutputPort
        """

        for i, p in enumerate(self.digital_input_ports):
            if port == p.info[1]:
                return p
        self._use_digital_input_port(port)
        return self.digital_input_ports[-1]

    def __contains__(self, port) -> bool:
        for p in self.ports:
            if port == p:
                return True
        return False

    def _use_analog_input_port(self, port_num: int, offset: float = 0.0):
        """Adds an instance of  AnalogInputPort to configuration if it is not used

        :param port_num: port number
        :type port_num: int
        :param offset: voltage offset for this port, defaults to 0.0
        :type offset: float, optional
        """
        if port_num not in self.dict["analog_inputs"]:
            self.dict["analog_inputs"][port_num] = {"offset": offset}
            self.analog_input_ports += [AnalogInputPort(self.name, port_num, offset=offset)]

    def _use_analog_output_port(self, port_num: int, offset: float = 0.0):
        """Adds an instance of  AnalogOutputPort to configuration if it is not used

        :param port_num: port number
        :type port_num: int
        :param offset: voltage offset for this port, defaults to 0.0
        :type offset: float, optional
        """
        if port_num not in self.dict["analog_outputs"]:
            self.dict["analog_outputs"][port_num] = {"offset": offset}
            self.analog_output_ports += [AnalogOutputPort(self.name, port_num, offset=offset)]

    def _use_digital_output_port(self, port_num: int):
        """Adds an instance of  DigitalOutputPort to configuration if it is not used

        :param port_num: port number
        :type port_num: int
        """
        if port_num not in self.dict["digital_outputs"]:
            self.dict["digital_outputs"][port_num] = {}
            self.digital_output_ports += [DigitalOutputPort(self.name, port_num)]

    def _use_digital_input_port(self, port_num: int):
        """Adds an instance of  DigitalInputPort to configuration if it is not used

        :param port_num: port number
        :type port_num: int
        """
        if port_num not in self.dict["digital_inputs"]:
            self.dict["digital_inputs"][port_num] = {}
            self.digital_input_ports += [DigitalInputPort(self.name, port_num)]

    @property
    def ports(self):
        """Returns the set of all used ports in this configuration

        :return:
        :rtype: Set
        """
        return self.digital_input_ports + self.digital_output_ports + self.analog_input_ports + self.analog_output_ports


class ArbitraryWaveform(Waveform):
    def __init__(self, name: str, samples: List[float]):
        """An arbitrary waveform played by the QOP

        :param name: name of the played WF
        :type name: str
        :param samples: list of samples to be played (in the range -.5 to .5), specified at 1 sample per nanosecond
        :type samples: List[float]
        """
        super().__init__(name, {"type": "arbitrary", "samples": samples})

    @property
    def samples(self):
        return self.dict["samples"]

    @samples.setter
    def samples(self, samples: List[float]):
        self.dict["samples"] = samples


class ConstantWaveform(Waveform):
    def __init__(self, name: str, sample: float):
        """A constant amplitude WF

        :param name: name of the WF
        :type name: str
        :param sample: Amplitude of the WF in the range -.5 to .5
        :type sample: float
        """
        super().__init__(name, {"type": "constant", "sample": sample})

    @property
    def sample(self):
        return self.dict["sample"]

    @sample.setter
    def sample(self, samples: List[float]):
        self.dict["sample"] = samples


class DigitalWaveform(Waveform):
    def __init__(self, name: str, samples: List[DigitalSample]):
        """A digital waveform played from a digital output

        :param name: The name of the digital waveform, defaults to "ON"
        :type name: str
        :param samples: The digital waveforms, specified in the format [(1/0, dur),...] where dur is in nanoseconds.  defaults to [(1, 0)]
        :type samples: List[DigitalSample]
        """

        if isinstance(samples[0], DigitalSample):
            self._samples = samples
        else:
            self._samples = [DigitalSample(state, duration) for (state, duration) in samples]
        dic = dict()
        dic["samples"] = [(ds.state, ds.duration) for ds in self._samples]
        super(DigitalWaveform, self).__init__(name, dic)

    @property
    def samples(self):
        return self._samples

    @samples.setter
    def samples(self, samples: Union[List[Tuple[int, int]], List[DigitalSample]]):
        if isinstance(samples[0], DigitalSample):
            self._samples = samples
        else:
            self._samples = [DigitalSample(state, duration) for (state, duration) in samples]
        self.dict["samples"] = [(ds.state, ds.duration) for ds in self._samples]


class MeasurePulse(Pulse):
    def __init__(self, name: str, wfs: List[Waveform], length: int, **kwargs):
        """A pulse which can be used by a QUA measure() command

        :param name: name for this pulse
        :type name: str
        :param wfs: waveforms to be played during measurement
        :type wfs: List[Waveform]
        :param length: duration of the measurment pulse
        :type length: int, optional
        """

        integration_weights: List[Weights] = kwargs.get("integration_weights", None)
        digital_marker: Optional[DigitalWaveform] = kwargs.get("digital_marker", None)
        super().__init__(name, wfs, "measurement", length, digital_marker)
        if integration_weights is None:
            integration_weights = []
        self.integration_weights = integration_weights

    def add(
        self,
        objs: Union[Weights, DigitalWaveform, List[Union[Weights, DigitalWaveform]]],
    ):
        if isinstance(objs, list):
            for obj in objs:
                self._add(obj)
        else:
            self._add(objs)
        return self

    def _add(self, w: Union[Weights, IntegrationWeights, DigitalWaveform]):
        if isinstance(w, IntegrationWeights):
            w = Weights(w)

        if isinstance(w, Weights):
            self.integration_weights.append(w)
        elif isinstance(w, DigitalWaveform):
            self.digital_marker = w


class ControlPulse(Pulse):
    def __init__(self, name: str, wfs: List[Waveform], length: int, **kwargs):
        """A pulse which can be used by a QUA play() command
        :return:
        :rtype: ControlPulse
        """
        digital_marker: Optional[DigitalWaveform] = kwargs.get("digital_marker", None)
        super().__init__(name, wfs, "control", length, digital_marker)

    def add(self, w: DigitalWaveform):
        self.digital_marker = w
        return self


class Mixer(ConfigBuilderElement):
    def __init__(self, name: str, data: List[MixerData] = None):
        """A microwave mixer

        :param name: name for this mixer
        :type name: str
        :param data: a list of mixer data with correction matrices for every pair of IF and LO
        :type data: List[MixerData]

        """
        super(Mixer, self).__init__(name)
        if data is None:
            data = []
        self.dict = []
        for e in data:
            self.add(e)

    def add(self, data: MixerData):
        self.dict.append(
            {
                "intermediate_frequency": data.intermediate_frequency,
                "lo_frequency": data.lo_frequency,
                "correction": list(data.correction.data[0] + data.correction.data[1]),
            }
        )
        return self


class Oscillator(ConfigBuilderElement):
    def __init__(self, name: str, intermediate_frequency: int, lo_frequency: int, mixer: str):
        """An internal oscillator.
        :param name: name of the oscillator
        :type name: str
        :param intermediate_frequency: intermediate frequency of the oscillator
        :type intermediate_frequency: int
        :param lo_frequency: local oscillator frequency
        :type lo_frequency: int
        :param mixer: mixer name
        :type mixer: str
        """
        super(Oscillator, self).__init__(name)
        self.dict = dict()
        self.dict["intermediate_frequency"] = intermediate_frequency
        self.dict["lo_frequency"] = lo_frequency
        self.dict["mixer"] = mixer

    @property
    def lo_frequency(self):
        return self.dict["lo_frequency"]

    @lo_frequency.setter
    def lo_frequency(self, lo_frequency: int):
        self.dict["lo_frequency"] = lo_frequency

    @property
    def intermediate_frequency(self):
        return self.dict["intermediate_frequency"]

    @intermediate_frequency.setter
    def intermediate_frequency(self, intermediate_frequency: int):
        self.dict["intermediate_frequency"] = intermediate_frequency

    @property
    def mixer(self):
        return self.dict["mixer"]

    @mixer.setter
    def mixer(self, mixer: str):
        self.dict["mixer"] = mixer


class Element(ConfigBuilderElement):
    def __init__(
        self,
        name: str,
        element_analog_inputs: Optional[List[AnalogOutputPort]] = None,
        element_analog_outputs: Optional[List[AnalogInputPort]] = None,
        element_digital_inputs: Optional[List[DigitalOutputPort]] = None,
        intermediate_frequency: Optional[int] = None,
        lo_frequency: Optional[int] = None,
        **kwargs
    ):
        """A quantum element in a configuration

        :param name: element name
        :type name: int
        :param element_analog_inputs: AnalogOutputPort(s) associated with this element, defaults to []
        :type element_analog_inputs: List[AnalogOutputPort], optional
        :param element_analog_outputs: AnalogInputPort(s) associated with this element, defaults to []
        :type element_analog_outputs: List[AnalogInputPort], optional
        :param element_digital_inputs: DigitalOutputPort(s) associated with this element, defaults to []
        :type element_digital_inputs: List[DigitalOutputPort], optional
        :param intermediate_frequency: intermediate frequency associated wit this element, defaults to None
        :type intermediate_frequency: int, optional
        :param lo_frequency: LO frequency associated wit this element, defaults to None
        :type lo_frequency: int, optional
        """
        super(Element, self).__init__(name)

        mixer: Optional[Mixer] = kwargs.get("mixer", None)
        pulses: Optional[List[Pulse]] = kwargs.get("pulses", None)
        operations: Optional[Dict[str, str]] = kwargs.get("digital_marker", None)

        analog_input_ports = _warn_if_not_none(kwargs, "analog_input_ports")
        if analog_input_ports is not None:
            element_analog_inputs = analog_input_ports

        analog_output_ports = _warn_if_not_none(kwargs, "analog_output_ports")
        if analog_output_ports is not None:
            element_analog_outputs = analog_output_ports

        digital_input_ports = _warn_if_not_none(kwargs, "digital_input_ports")
        if digital_input_ports is not None:
            element_digital_inputs = digital_input_ports

        element_digital_outputs = []
        digital_output_ports = _warn_if_not_none(kwargs, "digital_output_ports")
        if digital_output_ports is not None:
            element_digital_outputs = digital_output_ports

        self.dict = dict()
        self.pulses = []
        if pulses is not None:
            self.pulses = pulses
        self.element_analog_inputs = [] if element_analog_inputs is None else element_analog_inputs
        self.element_analog_outputs = [] if element_analog_outputs is None else element_analog_outputs
        self.element_digital_outputs = [] if element_digital_outputs is None else element_digital_outputs
        self.element_digital_inputs = [] if element_digital_inputs is None else element_digital_inputs

        assert len(self.element_analog_inputs) <= 2
        self.type = "singleInput" if len(self.element_analog_inputs) == 1 else "mixInputs"
        self.mixer: Mixer = mixer

        if len(self.element_analog_inputs) > 0:
            if self.type == "singleInput":
                port = self.element_analog_inputs[0]
                self.dict["singleInput"] = dict()
                self.dict["singleInput"]["port"] = port.info
            elif self.type == "mixInputs":
                I = self.element_analog_inputs[0]
                Q = self.element_analog_inputs[1]
                self.dict["mixInputs"] = dict()
                self.dict["mixInputs"]["I"] = I.info
                self.dict["mixInputs"]["Q"] = Q.info
                self.dict["mixInputs"]["lo_frequency"] = lo_frequency
        if len(self.element_analog_outputs) > 0:
            self.dict["outputs"] = dict()
            for i, port in enumerate(self.element_analog_outputs):
                self.dict["outputs"]["out" + str(i + 1)] = port.info
        self.dict["operations"] = dict()
        if operations is not None:
            self.dict["operations"] = operations
        self.dict["intermediate_frequency"] = intermediate_frequency
        # self.dict["outputPulseParameters"] = dict()
        if len(self.element_digital_inputs) > 0:
            self.dict["digitalInputs"] = dict()

            for port in self.element_digital_inputs:
                self.dict["digitalInputs"]["in" + str(port.port_id)] = {
                    "port": port.info,
                    "delay": 0,
                    "buffer": 0,
                }
        self.oscillator = None

    def set_digital_input_delay(self, port_id: int, val: int):
        if "in" + str(port_id) in self.dict["digitalInputs"].keys():
            self.dict["digitalInputs"]["in" + str(port_id)]["delay"] = val
        else:
            raise ConfigurationError("digital input port must be set first")

    def set_digital_input_buffer(self, port_id: int, val: int):
        if "in" + str(port_id) in self.dict["digitalInputs"].keys():
            self.dict["digitalInputs"]["in" + str(port_id)]["buffer"] = val
        else:
            raise ConfigurationError("digital input port must be set first")

    def _set_mixer(self, mix: Mixer):
        assert "mixInputs" in self.dict.keys()
        self.dict["mixInputs"]["mixer"] = mix.name
        self.mixer = mix

    def _set_oscillator(self, osc: Oscillator):
        self.dict["oscillator"] = osc.name
        self.oscillator = osc

    @property
    def has_lo_frequency(self):
        return "mixInputs" in self.dict.keys()

    @property
    def lo_frequency(self):
        assert "mixInputs" in self.dict.keys()
        return self.dict["mixInputs"]["lo_frequency"]

    @lo_frequency.setter
    def lo_frequency(self, lo_frequency: int):
        assert "mixInputs" in self.dict.keys()
        self.dict["mixInputs"]["lo_frequency"] = lo_frequency

    @property
    def intermediate_frequency(self):
        return self.dict["intermediate_frequency"]

    @intermediate_frequency.setter
    def intermediate_frequency(self, omega_IF: int):
        """Set the IF associated with this element

        :param omega_IF:
        :type omega_IF: int
        """
        assert self.type == "mixInputs"
        self.dict["intermediate_frequency"] = omega_IF

    @property
    def operations(self):
        return self.dict["operations"]

    def _add_operation(self, op: Operation):
        self.dict["operations"][op.name] = op.pulse.name
        self.pulses.append(op.pulse)

    def _add(self, obj: Union[Operation, Mixer, ControlPulse, MeasurePulse, Oscillator]):
        """A method to add components to an element

        :param obj: The component to be added
        :type obj: [type]
        :raises ConfigurationError: [description]
        """
        if isinstance(obj, Operation):
            self._add_operation(obj)
        elif isinstance(obj, Mixer):
            self._set_mixer(obj)
        elif isinstance(obj, ControlPulse) or isinstance(obj, MeasurePulse):
            self.dict["operations"][obj.name] = obj.name
            self.pulses.append(obj)
        elif isinstance(obj, Oscillator):
            self._set_oscillator(obj)
        else:
            raise ConfigurationError("Adding unsupported object")

    def add(self, objs: Union[Operation, Mixer, List[Union[Operation, Mixer]]]):
        if isinstance(objs, list):
            for obj in objs:
                self._add(obj)
        else:
            self._add(objs)
        return self

    @property
    def operation_names(self):
        return self.dict["operations"].keys()

    @property
    def pulse_names(self):
        return set([p.name for p in self.pulses])

    @property
    def waveform_names(self):
        names = []
        for pl in self.pulses:
            names += pl.waveform_names
        return set(names)

    @property
    def ports(self):
        return self.analog_input_ports + self.analog_output_ports + self.digital_output_ports + self.digital_input_ports

    @property
    def has_signal_threshold(self):
        return "outputPulseParameters" in self.dict and "signalThreshold" in self.dict["outputPulseParameters"].keys()

    @property
    def signal_threshold(self):
        if "signalThreshold" in self.dict["outputPulseParameters"].keys():
            return self.dict["outputPulseParameters"]["signalThreshold"]
        else:
            raise ConfigurationError("set the signal threshold")

    @signal_threshold.setter
    def signal_threshold(self, val: int):
        if "outputPulseParameters" not in self.dict.keys():
            self.dict["outputPulseParameters"] = dict()
        self.dict["outputPulseParameters"]["signalThreshold"] = val

    @property
    def has_signal_polarity(self):
        return "outputPulseParameters" in self.dict and "signalPolarity" in self.dict["outputPulseParameters"].keys()

    @property
    def signal_polarity(self):
        if "signalPolarity" in self.dict["outputPulseParameters"].keys():
            return self.dict["outputPulseParameters"]["signalPolarity"]
        else:
            raise ConfigurationError("set the signal polarity")

    @signal_polarity.setter
    def signal_polarity(self, val: str):
        if "outputPulseParameters" not in self.dict.keys():
            self.dict["outputPulseParameters"] = dict()
        self.dict["outputPulseParameters"]["signalPolarity"] = val

    @property
    def has_derivative_threshold(self):
        return (
            "outputPulseParameters" in self.dict and "derivativeThreshold" in self.dict["outputPulseParameters"].keys()
        )

    @property
    def derivative_threshold(self):
        if "derivativeThreshold" in self.dict["outputPulseParameters"].keys():
            return self.dict["outputPulseParameters"]["derivativeThreshold"]
        else:
            raise ConfigurationError("set the derivative threshold")

    @derivative_threshold.setter
    def derivative_threshold(self, val: int):
        if "outputPulseParameters" not in self.dict.keys():
            self.dict["outputPulseParameters"] = dict()
        self.dict["outputPulseParameters"]["derivativeThreshold"] = val

    @property
    def has_derivative_polarity(self):
        return (
            "outputPulseParameters" in self.dict and "derivativePolarity" in self.dict["outputPulseParameters"].keys()
        )

    @property
    def derivative_polarity(self):
        if "derivativePolarity" in self.dict["outputPulseParameters"].keys():
            return self.dict["outputPulseParameters"]["derivativePolarity"]
        else:
            raise ConfigurationError("set the derivative polarity")

    @derivative_polarity.setter
    def derivative_polarity(self, val: str):
        if "outputPulseParameters" not in self.dict.keys():
            self.dict["outputPulseParameters"] = dict()
        self.dict["outputPulseParameters"]["derivativePolarity"] = val

    @property
    def thread(self):
        return self.dict["thread"]

    @thread.setter
    def thread(self, thread: str):
        self.dict["thread"] = thread


class MeasureElement(Element):
    def __init__(
        self,
        name: str,
        element_analog_inputs: List[AnalogOutputPort],
        element_analog_outputs: List[AnalogInputPort],
        element_digital_inputs: Optional[List[DigitalOutputPort]] = None,
        intermediate_frequency: Optional[int] = None,
        lo_frequency: Optional[int] = None,
        **kwargs
    ):
        """A quantum element set for performing measurment

        :param name: [description]
        :type name: int
        :param element_analog_inputs: [description]
        :type element_analog_inputs: List[AnalogOutputPort]
        :param element_analog_outputs: [description]
        :type element_analog_outputs: List[AnalogInputPort]
        :param element_digital_inputs: [description], defaults to []
        :type element_digital_inputs: List[DigitalOutputPort], optional
        :param intermediate_frequency: intermediate frequency associated wit this element, defaults to None
        :type intermediate_frequency: int, optional
        :param lo_frequency: [description], defaults to None
        :type lo_frequency: int, optional
        """

        time_of_flight: Optional[int] = kwargs.get("time_of_flight", 0)
        smearing: Optional[int] = kwargs.get("smearing", 0)
        mixer: Optional[Mixer] = kwargs.get("mixer", None)
        pulses: Optional[List[Pulse]] = kwargs.get("pulses", None)
        operations: Optional[Dict[str, str]] = kwargs.get("operations", None)

        analog_input_ports = _warn_if_not_none(kwargs, "analog_input_ports")
        if analog_input_ports is not None:
            element_analog_inputs = analog_input_ports

        analog_output_ports = _warn_if_not_none(kwargs, "analog_output_ports")
        if analog_output_ports is not None:
            element_analog_outputs = analog_output_ports

        digital_input_ports = _warn_if_not_none(kwargs, "digital_input_ports")
        if digital_input_ports is not None:
            element_digital_inputs = digital_input_ports

        element_digital_outputs = []
        digital_output_ports = _warn_if_not_none(kwargs, "digital_output_ports")
        if digital_output_ports is not None:
            element_digital_outputs = digital_output_ports

        super().__init__(
            name,
            element_analog_inputs,
            element_analog_outputs,
            element_digital_inputs,
            intermediate_frequency,
            lo_frequency,
            element_digital_outputs=element_digital_outputs,
            mixer=mixer,
            pulses=pulses,
            operations=operations,
        )
        self.dict["time_of_flight"] = time_of_flight
        self.dict["smearing"] = smearing

    @property
    def time_of_flight(self):
        return self.dict["time_of_flight"]

    @time_of_flight.setter
    def time_of_flight(self, tof: int):
        self.dict["time_of_flight"] = tof

    @property
    def smearing(self):
        return self.dict["smearing"]

    @smearing.setter
    def smearing(self, smearing: int):
        self.dict["smearing"] = smearing


class PiecewiseConstantIntegrationWeights(IntegrationWeights):
    def __init__(
        self,
        name: str,
        cosines: List[Tuple[float, int]],
        sines: List[Tuple[float, int]],
    ):
        """Piece-wise constant integration weights used in integration and demodulation
        :param name: name for the weights vector
        :type name: str
        :param cosines: values of the cosine weight and the respective durations
        :type cosines: List[Tuple[float, int]]
        :param sines: values of the sine weight and the respective durations
        :type sines: List[Tuple[float, int]]
        """
        self._cosines = cosines
        self._sines = sines
        tot_duration_cos = 0
        for e in cosines:
            tot_duration_cos += e[1]
        tot_duration_sin = 0
        for e in sines:
            tot_duration_sin += e[1]
        assert tot_duration_sin == tot_duration_cos, "Total duration of the sine and cosine weights must be same"

        super(PiecewiseConstantIntegrationWeights, self).__init__(
            name,
            cosines,
            sines,
        )


class ConstantIntegrationWeights(IntegrationWeights):
    def __init__(self, name: str, cosine: float, sine: float, duration: int):
        """A constant integration weight used in integration and demodulation
        :param name: name for the weights vector
        :type name: str
        :param cosine: value of the cosine vector
        :type cosine: float
        :param sine: value of the sine vector
        :type sine: float
        :param duration: duration of the read out pulse
        :type duration: int
        """
        self._duration = duration
        self._cosine = cosine
        self._sine = sine
        super(ConstantIntegrationWeights, self).__init__(name, [(cosine, duration)], [(sine, duration)])

    @property
    def cosine(self):
        return self.dict["cosine"]

    @cosine.setter
    def cosine(self, v: float):
        self.dict["cosine"] = [(v, self._duration)]

    @property
    def sine(self):
        return self.dict["sine"]

    @sine.setter
    def sine(self, v: float):
        self.dict["sine"] = [(v, self._duration)]


class ArbitraryIntegrationWeights(IntegrationWeights):
    def __init__(self, name: str, cosine: List[float], sine: List[float]):
        """A vector of integration weights used in integration and demodulation
        :param name: name for the weights vector
        :type name: str
        :param cosine: a list of weights to be used with the cosine demodulation element, specified at 1 sample per 4 nanoseconds
        :type cosine: List
        :param sine: a list of weights to be used with the sine demodulation element, specified at 1 sample per 4 nanoseconds
        :type sine: List
        """
        super().__init__(name, cosine, sine)
        assert len(sine) == len(cosine)

    @property
    def cosine(self):
        return self.dict["cosine"]

    @cosine.setter
    def cosine(self, w: List[float]):
        self.dict["cosine"] = w

    @property
    def sine(self):
        return self.dict["sine"]

    @sine.setter
    def sine(self, w: List[float]):
        self.dict["sine"] = w


class ElementCollection(ConfigBuilderElement):
    def __init__(self, name: str):
        """A named collection of Elements

        It also a useful base class for building components.
        """
        super(ElementCollection, self).__init__(name)

    @abc.abstractmethod
    def get_elements(self) -> List[ConfigBuilderElement]:
        pass

    @property
    def ports(self):
        ports = []
        for elm in self.get_elements():
            ports += elm.ports
        return ports

    @property
    def waveform_names(self):
        names = []
        for elm in self.get_elements():
            names += elm.waveform_names
        return names

    @property
    def operation_names(self):
        names = []
        for elm in self.get_elements():
            names += elm.operation_names
        return names

    @property
    def pulse_names(self):
        names = []
        for elm in self.get_elements():
            names += elm.pulse_names
        return names


class ReadoutResonator(ElementCollection):
    def __init__(
        self,
        name: str,
        outputs: List[AnalogOutputPort],
        inputs: List[AnalogInputPort],
        intermediate_frequency: int,
        **kwargs
    ):
        """
        A Microwave readout resonator

        :param name: A name for this resonator
        :type name: str
        :param outputs: A pair of output ports
        :type outputs: List[AnalogOutputPort]
        :param inputs: A pair of input ports
        :type inputs: List[AnalogInputPort]
        :param intermediate_frequency: The intermediate frequency
        :type intermediate_frequency: int
        """

        operations: Optional[List[Operation]] = kwargs.get("operations", None)
        time_of_flight: Optional[int] = kwargs.get("time_of_flight", 0)
        smearing: Optional[int] = kwargs.get("smearing", 0)
        mixer: Optional[Mixer] = kwargs.get("mixer", None)
        lo_frequency: Optional[int] = kwargs.get("lo_frequency", None)

        super(ReadoutResonator, self).__init__(name=name)
        if len(outputs) != 2:
            raise ConfigurationError("Number of output ports is not equal to 2")
        if len(inputs) != 2:
            raise ConfigurationError("Number of input ports is not equal to 2")
        self.drive_name = name
        self.drive_outputs = outputs
        self.drive_inputs = inputs
        self.drive_operations = []
        if operations is not None:
            self.drive_operations = operations
        self.drive_intermediate_frequency = intermediate_frequency
        self.drive_time_of_flight = time_of_flight

        self.drive_mixer = mixer
        self.drive_lo_frequency = lo_frequency
        self.drive_smearing = smearing

    def get_elements(self) -> List[ConfigBuilderElement]:
        drive = MeasureElement(
            self.drive_name,
            self.drive_outputs,
            self.drive_inputs,
            intermediate_frequency=self.drive_intermediate_frequency,
            time_of_flight=self.drive_time_of_flight,
            mixer=self.drive_mixer,
            lo_frequency=self.drive_lo_frequency,
            smearing=self.drive_smearing,
        )
        for op in self.drive_operations:
            drive.add(op)

        if self.drive_inputs:
            for i, port in enumerate(self.drive_inputs):
                drive.dict["outputs"]["out" + str(i + 1)] = port.info

        drive.dict["mixInputs"]["I"] = self.drive_outputs[0].info
        drive.dict["mixInputs"]["Q"] = self.drive_outputs[1].info

        if self.drive_mixer:
            drive.dict["mixInputs"]["mixer"] = self.drive_mixer.name

        return [drive]

    @property
    def intermediate_frequency(self):
        return self.drive_intermediate_frequency

    @intermediate_frequency.setter
    def intermediate_frequency(self, if_freq: int):
        """Set an IF for the resonator

        :param if_freq: A frequency in MHz
        :type if_freq: int
        """
        self.drive_intermediate_frequency = if_freq

    def add(self, obj: Union[Operation, List[Operation]]):
        """Add list of operations associated with the Transmon

        :param obj: The list of operations
        :type obj: list or an instance of Operation
        """
        if isinstance(obj, list):
            for o in obj:
                self._add(o)
        else:
            self._add(obj)
        return self

    def _add(self, op: Union[Operation, MeasurePulse]):
        """Add operation associated with the Readout resonator

        :param obj: The operation
        :type obj: Operation

        """
        if isinstance(op, Operation):
            if len(op.pulse.wfs) == 2:
                self.drive_operations.append(op)
            else:
                raise ConfigurationError("adding unsupported object")
        elif isinstance(op, MeasurePulse):
            if len(op.wfs) == 2:
                self.drive_operations.append(op)
            else:
                raise ConfigurationError("adding unsupported object")
        else:
            raise ConfigurationError("adding unsupported object")

    @property
    def time_of_flight(self):
        return self.drive_time_of_flight

    @time_of_flight.setter
    def time_of_flight(self, tof: int):
        self.drive_time_of_flight = tof

    @property
    def smearing(self):
        return self.drive_smearing

    @smearing.setter
    def smearing(self, t: int):
        self.drive_smearing = t

    @property
    def lo_frequency(self):
        return self.drive_lo_frequency

    @lo_frequency.setter
    def lo_frequency(self, lo_frequency: int):
        self.drive_lo_frequency = lo_frequency

    @property
    def input_ports(self):
        return self.drive_inputs

    @input_ports.setter
    def input_ports(self, ports: List[AnalogInputPort]):
        assert len(ports) == 2
        self.drive_inputs = ports

    @property
    def output_ports(self):
        return self.drive_outputs

    @output_ports.setter
    def output_ports(self, ports: List[AnalogOutputPort]):
        assert len(ports) == 2
        self.drive_outputs = ports

    @property
    def mixer(self):
        return self.drive_mixer

    @mixer.setter
    def mixer(self, mixer: Mixer):
        self.drive_mixer = mixer


class Transmon(ElementCollection):
    def __init__(self, name: str, I: AnalogOutputPort, Q: AnalogOutputPort, intermediate_frequency: int, **kwargs):
        """A superconducting transmon qubit

        :param name: A name for this transmon
        :type name: str
        :param I: I output port
        :type I: AnalogOutputPort
        :param Q: Q output port
        :type Q: AnalogOutputPort
        :param intermediate_frequency: intermediate frequency of the upconversion IQ mixer in MHz
        :type intermediate_frequency: int
        """

        mixer: Optional[Mixer] = kwargs.get("mixer", None)
        lo_frequency: Optional[int] = kwargs.get("lo_frequency", None)
        operations: Optional[List[Operation]] = kwargs.get("operations", None)

        super(Transmon, self).__init__(name=name)
        self.drive_I = I
        self.drive_Q = Q
        self.drive_intermediate_frequency = intermediate_frequency

        self.drive_operations = []
        if operations is not None:
            self.drive_operations = operations
        self.drive_intermediate_frequency = intermediate_frequency
        self.drive_mixer = mixer
        self.drive_lo_frequency = lo_frequency

    def get_elements(self) -> List[ConfigBuilderElement]:
        drive = Element(
            self.name,
            [self.drive_I, self.drive_Q],
            intermediate_frequency=self.drive_intermediate_frequency,
            mixer=self.drive_mixer,
            lo_frequency=self.drive_lo_frequency,
        )
        drive.dict["mixInputs"]["I"] = self.drive_I.info
        drive.dict["mixInputs"]["Q"] = self.drive_Q.info

        for op in self.drive_operations:
            drive.add(op)

        if self.drive_mixer:
            drive.dict["mixInputs"]["mixer"] = self.drive_mixer.name

        return [drive]

    @property
    def I(self):
        return self.drive_I

    @I.setter
    def I(self, p: AnalogOutputPort):
        self.drive_I = p

    @property
    def Q(self):
        return self.drive_Q

    @Q.setter
    def Q(self, p: AnalogOutputPort):
        self.drive_Q = p

    @property
    def lo_frequency(self):
        return self.drive_lo_frequency

    @lo_frequency.setter
    def lo_frequency(self, lo_frequency: int):
        self.drive_lo_frequency = lo_frequency

    @property
    def intermediate_frequency(self):
        return self.drive_intermediate_frequency

    @intermediate_frequency.setter
    def intermediate_frequency(self, if_freq: int):
        """
            Set an IF for the transmon
        :param if_freq: A frequency in MHz
        :type if_freq: int
        """
        self.drive_intermediate_frequency = if_freq

    def add(self, op: Union[Operation, List[Operation]]):
        """
            Add list of operations associated with the Transmon
        :param op: The operations
        :type op: list or an instance of Operation
        """
        if isinstance(op, list):
            for o in op:
                self._add(o)
        else:
            self._add(op)
        return self

    def _add(self, op: Union[Operation, ControlPulse]):
        if isinstance(op, Operation):
            if len(op.pulse.wfs) == 2:
                self.drive_operations.append(op)
            else:
                raise ConfigurationError("adding unsupported object")
        elif isinstance(op, ControlPulse):
            if len(op.wfs) == 2:
                self.drive_operations.append(op)
            else:
                raise ConfigurationError("adding unsupported object")
        else:
            raise ConfigurationError("adding unsupported object")

    @property
    def mixer(self):
        return self.drive_mixer

    @mixer.setter
    def mixer(self, mixer: Mixer):
        self.drive_mixer = mixer


class FluxTunableTransmon(Transmon):
    def __init__(
        self,
        name: str,
        I: AnalogOutputPort,
        Q: AnalogOutputPort,
        flux_port: AnalogOutputPort,
        intermediate_frequency: int,
        **kwargs
    ):
        """A flux tuneable superconducting transmon qubit

        :param name: A name for this transmon
        :type name: str
        :param I: I output port
        :type I: AnalogOutputPort
        :param Q: Q output port
        :type Q: AnalogOutputPort
        :param flux_port:  Flux line output Port
        :type flux_port: AnalogOutputPort
        :param intermediate_frequency: intermediate frequency of the upconversion IQ mixer in MHz
        :type intermediate_frequency: int
        """

        mixer: Optional[Mixer] = kwargs.get("mixer", None)
        lo_frequency: Optional[int] = kwargs.get("lo_frequency", None)
        operations: Optional[List[Operation]] = kwargs.get("operations", None)
        flux_operations: Optional[List[Operation]] = kwargs.get("flux_operations", None)

        super(FluxTunableTransmon, self).__init__(
            name,
            I,
            Q,
            intermediate_frequency,
            mixer=mixer,
            lo_frequency=lo_frequency,
            operations=operations,
        )
        self.flux_port = flux_port
        self.flux_operations = []
        if flux_operations is not None:
            self.flux_operations = flux_operations

    def get_elements(self) -> List[ConfigBuilderElement]:
        els = super().get_elements()

        el = Element(self.name + "_flux_line", [self.flux_port])
        for op in self.flux_operations:
            el.add(op)

        els.append(el)
        return els

    def _add(self, op: Union[Operation, ControlPulse]):
        if isinstance(op, Operation):
            if len(op.pulse.wfs) == 2:
                super(FluxTunableTransmon, self)._add(op)
            elif len(op.pulse.wfs) == 1:
                self.flux_operations.append(op)
            else:
                raise ConfigurationError("Can not add a pulse with {0} waveforms".format(len(op.pulse.wfs)))
        elif isinstance(op, ControlPulse):
            if len(op.wfs) == 2:
                super(FluxTunableTransmon, self)._add(op)
            elif len(op.wfs) == 1:
                self.flux_operations.append(op)
            else:
                raise ConfigurationError("Can not add a pulse with {0} waveforms".format(len(op.wfs)))

        else:
            raise ConfigurationError("adding unsupported object of type {}".format(type(op)))


class Coupler(ElementCollection):
    def __init__(self, name: str, port: AnalogOutputPort, **kwargs):
        """Element driving a coupler between two transmons
        :param name: name of the coupler
        :type name: str
        :param port: analog output port driving the coupler
        :type port: AnalogOutputPort
        """
        operations: Optional[List[Operation]] = kwargs.get("operations", None)
        super(Coupler, self).__init__(name=name)
        self._port = port
        self.operations = []
        if operations is not None:
            self.operations = operations

    def get_elements(self) -> List[ConfigBuilderElement]:
        coupler = Element(self.name, analog_input_ports=[self._port])
        for op in self.operations:
            coupler.add(op)
        return [coupler]

    def _add(self, op: Operation):
        if len(op.pulse.wfs) == 1:
            self.operations.append(op)
        else:
            raise ConfigurationError("Can not add a pulse with {0} waveforms".format(len(op.pulse.wfs)))

    def add(self, op: Union[List[Operation], Operation]):
        if isinstance(op, list):
            for o in op:
                self._add(o)
        else:
            self._add(op)
        return self

    @property
    def port(self):
        return self._port

    @port.setter
    def port(self, p: AnalogOutputPort):
        self._port = p



import logging
from contextlib import contextmanager
from time import sleep

from qm import QuantumMachine
from qm import QuantumMachinesManager

msg = (
    "A quantum machine cannot be opened because an existing quantum machine, using the same ports, is currently "
    "running a program. Please close the currently open quantum machine."
)


class BusyFilter(logging.Filter):
    def filter(self, record):
        return not ("already uses" in record.getMessage() or msg in record.getMessage())


@contextmanager
def qm_session(qmm: QuantumMachinesManager, config: dict, timeout: int = 100) -> QuantumMachine:
    """
    This context manager allows a user to _try_ to
    open a quantum machine, and if it is not possible since its resources are currently in use,
    wait for them to free up. This is done by repeatedly polling the QM manager.

    Once they are freed, execution will start automatically.
    Once the context manager code block is exited, the QM will close automatically, freeing it up
    for another user.

    :param qmm: A QM manager from which to open QM
    :param config: a QUA config that will be supplied to `open_qm`
    :param timeout: time in seconds to wait for resources to free up before raising exception

    """
    if not timeout > 0:
        raise ValueError(f"timeout={timeout} but must be positive")
    qm_log = logging.getLogger("qm")
    filt = BusyFilter()
    is_busy = True
    printed = False
    time = 0
    while is_busy and time < timeout:
        try:
            qm_log.addFilter(filt)
            qm = qmm.open_qm(config, close_other_machines=False)
            # qm_log.warning(f"QOP is busy. Waiting for it to free up for {timeout}s...")
        except Exception as e:
            if (
                hasattr(e, "errors")
                and isinstance(e.errors, list)
                and len(e.errors) > 1
                and isinstance(e.errors[1], tuple)
                and len(e.errors[1]) > 2
                and msg in e.errors[1][2]
            ):
                if not printed:
                    qm_log.warning(f"QOP is busy. Waiting for it to free up for {timeout}s...")
                    printed = True
                sleep(0.2)
                time += 0.2
            else:
                raise Exception from e
        else:
            is_busy = False
            qm_log.info("Opening QM")
        finally:
            qm_log.removeFilter(filt)
    if is_busy and time >= timeout:
        qm_log.warning(f"While waiting for QOP to free, reached timeout: {timeout}s")
        raise TimeoutError(f"While waiting for QOP to free, reached timeout: {timeout}s")
    try:
        yield qm
        while qm.get_running_job() is not None:
            sleep(0.2)
    except KeyboardInterrupt:
        pass
    finally:
        qm_log.info("Closing QM")
        qm.close()



from collections import UserDict as _UserDict
from copy import deepcopy as _deepcopy
from typing import List, Tuple
import json as _json
from pprint import pprint


def transform_negative_delays(config, create_new_config=False):
    """
    Adds the most negative delay to all the elements in the configuration.

    :param config: A valid QOP configuration.
    :param create_new_config: If true, returns a new copy of the configuration.
        If false, edits the existing configuration. Default is false.
    :returns: The edited configuration.
    """
    min_delay = 0
    if create_new_config:
        config = _deepcopy(config)
    cons = config.get("controllers")
    for con in cons:
        ports = cons.get(con).get("analog_outputs")
        for port in ports:
            min_delay = min(ports.get(port).get("delay", 0), min_delay)

    if min_delay < 0:
        for con in cons:
            ports = cons.get(con).get("analog_outputs")
            for port in ports:
                delay = ports.get(port).get("delay", 0)
                ports.get(port)["delay"] = delay - min_delay
    return config


class QuaConfig(_UserDict):
    def __init__(self, data):
        super().__init__(data)
        self._data_orig = _deepcopy(data)
        self.pulse_suffix = "_pulse"
        self.wf_suffix = "_wf"

    def print(self):
        """
        Print the config using pprint.
        """
        return pprint(self)

    def reset(self):
        """
        Reset to config to its initial state.
        """
        self.data = _deepcopy(self._data_orig)

    def dump(self, filename):
        with open(filename, "w") as fp:
            _json.dump(self.data, fp)

    def add_control_operation_iq(self, element: str, operation_name: str, wf_i: List[float], wf_q: List[float]):
        """
        Add or update a control operation to a mixed input element.

        :param element: name of the mixed input element to add the operation to. Must be defined in the configuration.
        :param operation_name: name of the operation to be added.
        :param wf_i: list of points defining the waveform for the 'I' quadrature. Must have the same length as wf_q.
        :param wf_q: list of points defining the waveform for the 'Q' quadrature. Must have the same length as wf_i.
        """

        # Check inputs
        if element not in self.data["elements"].keys():
            raise KeyError(f"The element '{element}' in not defined in the config.")
        if "mixInputs" not in self.data["elements"][element].keys():
            raise KeyError(f"Element '{element}' is not a mixInputs element.")
        if len(wf_i) != len(wf_q):
            raise ValueError("The 'I' and 'Q' waveforms must have the same length.")

        # Add operation
        pulse_name = element + "_" + operation_name + self.pulse_suffix
        wf_name = element + "_" + operation_name + self.wf_suffix
        self.data["elements"][element]["operations"][operation_name] = pulse_name
        # Add pulse
        self.data["pulses"][pulse_name] = {
            "operation": "control",
            "length": len(wf_i),
            "waveforms": {"I": wf_name + "_i", "Q": wf_name + "_q"},
        }
        # Add waveform
        self.update_waveforms(element, operation_name, (wf_i, wf_q))

    def add_control_operation_single(self, element: str, operation_name: str, wf: List[float]):
        """
        Add or update a control operation to a single input element.

        :param element: name of the single input element to add the operation to. Must be defined in the configuration.
        :param operation_name: name of the operation to be added.
        :param wf: list of points defining the waveform with a 1ns resolution.
        """

        # Check inputs
        if element not in self.data["elements"].keys():
            raise KeyError(f"The element '{element}' in not defined in the config.")
        if "singleInput" not in self.data["elements"][element].keys():
            raise KeyError(f"Element '{element}' is not a singleInput element.")

        # Add operation
        pulse_name = element + "_" + operation_name + self.pulse_suffix
        wf_name = element + "_" + operation_name + "_single" + self.wf_suffix
        self.data["elements"][element]["operations"][operation_name] = pulse_name
        # Add pulse
        self.data["pulses"][pulse_name] = {
            "operation": "control",
            "length": len(wf),
            "waveforms": {"single": wf_name},
        }
        # Add waveform
        self.update_waveforms(element, operation_name, (wf,))

    def get_waveforms_from_op(self, element: str, operation_name: str) -> List or List[List]:
        """
        Get the waveforms corresponding to the given element and operation.

        :param element: name of the element to get the waveforms from. Must be defined in the config.
        :param operation_name: name of the operation to get the waveforms from. Must be defined in the config.
        :return: list of the corresponding waveforms.
        """

        # Check inputs
        if element not in self.data["elements"].keys():
            raise KeyError(f"The element '{element}' in not defined in the config.")
        if operation_name not in self.data["elements"][element]["operations"].keys():
            raise KeyError(
                f"The operation '{operation_name}' in not defined in the config for the element '{element}'."
            )

        # Get the waveforms
        pulse = self.get_pulse_from_op(element, operation_name)
        if "mixInputs" in self.data["elements"][element]:
            waveform_i = self.data["waveforms"][pulse["waveforms"]["I"]]
            if waveform_i["type"] == "arbitrary":
                waveform_i = waveform_i["samples"]
            else:
                waveform_i = [waveform_i["sample"]] * pulse["length"]

            waveform_q = self.data["waveforms"][pulse["waveforms"]["Q"]]
            if waveform_q["type"] == "arbitrary":
                waveform_q = waveform_q["samples"]
            else:
                waveform_q = [waveform_q["sample"]] * pulse["length"]
            return waveform_i, waveform_q
        else:
            waveform = self.data["waveforms"][pulse["waveforms"]["single"]]
            if waveform["type"] == "arbitrary":
                waveform = waveform["samples"]
            else:
                waveform = [waveform["sample"]] * pulse["length"]
            return waveform

    def get_pulse_from_op(self, element: str, operation_name: str) -> dict:
        """
        Get the pulse corresponding to the given element and operation.

        :param element: name of the element to get the waveforms from. Must be defined in the config.
        :param operation_name: name of the operation to get the waveforms from. Must be defined in the config.
        :return: the corresponding pulse as a dictionary.
        """

        # Check inputs
        if element not in self.data["elements"].keys():
            raise KeyError(f"The element '{element}' in not defined in the config.")
        if operation_name not in self.data["elements"][element]["operations"].keys():
            raise KeyError(
                f"The operation '{operation_name}' in not defined in the config for the element '{element}'."
            )

        return self.data["pulses"][self.data["elements"][element]["operations"][operation_name]]

    def update_op_amp(self, element: str, operation_name: str, amp: float, force_update: bool = False):
        """
        Update the operation amplitude.
        Can only access amplitude for a constant pulse of a single element.
        Updating the waveform may affect other elements if it is used in other operations.
        Set the force_update flag to True to update the waveform anyway.

        :param element: name of the element to get the waveforms from. Must be defined in the config.
        :param operation_name: name of the operation to get the waveforms from. Must be defined in the config.
        :param amp: float for the updated amplitude. Must be within the range [-0.5, 0.5) V.
        :param force_update: If True, the waveform will be updated even if it used in other operations.
        """

        # Check inputs
        if element not in self.data["elements"].keys():
            raise KeyError(f"The element '{element}' in not defined in the config.")
        if operation_name not in self.data["elements"][element]["operations"].keys():
            raise KeyError(
                f"The operation '{operation_name}' in not defined in the config for the element '{element}'."
            )
        if not -0.5 <= amp < 0.5:
            raise ValueError("The amplitude must be within the range [-0.5, 0.5) Volts.")
        pulse = self.get_pulse_from_op(element, operation_name)
        if "single" not in pulse["waveforms"].keys():
            raise KeyError("Can only access amplitude for a constant pulse of a single element")
        wf = pulse["waveforms"]["single"]
        # Check if the waveform is used in another pulse
        count = 0
        for key in self.data["pulses"].keys():
            if "single" in self.data["pulses"][key]["waveforms"]:
                if wf == self.data["pulses"][key]["waveforms"]["single"]:
                    count += 1
            elif "I" in self.data["pulses"][key]["waveforms"]:
                if wf == self.data["pulses"][key]["waveforms"]["I"]:
                    count += 1
            elif "Q" in self.data["pulses"][key]["waveforms"]:
                if wf == self.data["pulses"][key]["waveforms"]["Q"]:
                    count += 1
        if count > 1 and not force_update:
            raise Exception(
                "The updated waveform is used in other operations. To force the update, please set the force_update flag to True."
            )
        elif count == 0:
            raise Exception(f"The operation {operation_name} doesn't have a valid waveform.")
        # Update the waveform
        self.data["waveforms"][wf]["sample"] = amp

    def get_op_amp(self, element: str, operation_name: str) -> float:
        """
        Get the amplitude corresponding to the given element and constant operation.
        Can only access amplitude for a constant pulse of a single element.

        :param element: name of the element to get the waveforms from. Must be defined in the config.
        :param operation_name: name of the operation to get the waveforms from. Must be defined in the config.
        :return: the corresponding amplitude.
        """
        # Check inputs
        if element not in self.data["elements"].keys():
            raise KeyError(f"The element '{element}' in not defined in the config.")
        if operation_name not in self.data["elements"][element]["operations"].keys():
            raise KeyError(
                f"The operation '{operation_name}' in not defined in the config for the element '{element}'."
            )

        pulse = self.get_pulse_from_op(element, operation_name)
        try:
            return self.data["waveforms"][pulse["waveforms"]["single"]]["sample"]
        except KeyError:
            raise KeyError("Can only access amplitude for a constant pulse of a single element")

    def update_integration_weight(
        self,
        element: str,
        operation_name: str,
        iw_op_name: str,
        iw_cos: List,
        iw_sin: List,
        force_update: bool = False,
    ):
        """
        Update the cosine and sine parts of an integration weight for a given element and operation.

        :param element: name of the element to get the waveforms from. Must be defined in the config.
        :param operation_name: name of the operation to get the waveforms from. Must be defined in the config.
        :param iw_op_name: name of the integration weight to be updated. Must be defined in the config.
        :param iw_cos: values for the cosine part of integration weights defined as a list of tuples [(value, duration in ns)].
        :param iw_sin: values for the sine part of integration weights defined as a list of tuples [(value, duration in ns)].
        :param force_update: If True, the waveform will be updated even if it used in other operations.
        """
        # Check inputs
        if element not in self.data["elements"].keys():
            raise KeyError(f"The element '{element}' in not defined in the config.")
        if operation_name not in self.data["elements"][element]["operations"].keys():
            raise KeyError(
                f"The operation '{operation_name}' in not defined in the config for the element '{element}'."
            )
        if (
            iw_op_name
            not in self.data["pulses"][self.data["elements"][element]["operations"][operation_name]][
                "integration_weights"
            ].keys()
        ):
            raise KeyError(
                f"The integration weight '{iw_op_name}' in not associated to the pulse corresponding to the operation '{operation_name}'."
            )
        # Update iw
        pulse_name = self.data["elements"][element]["operations"][operation_name]
        iw_name = self.data["pulses"][pulse_name]["integration_weights"][iw_op_name]
        count = 0
        for key in self.data["pulses"]:
            if "integration_weights" in self.data["pulses"][key]:
                if iw_name in self.data["pulses"][key]["integration_weights"].values():
                    count += 1
        if count > 1 and not force_update:
            raise Exception(
                "The updated integration weights are used in other operations. To force the update, please set the force_update flag to True."
            )
        elif count == 0:
            raise Exception(f"The integration weights {iw_op_name} are not listed in the config integration weights.")

        self.data["integration_weights"][iw_name] = {"cosine": iw_cos, "sine": iw_sin}

    def copy_operation(self, element: str, operation_name: str, new_name: str):
        """
        Copy an existing operation and rename it. This function is used to add similar operations that differ
        only from their waveforms than can be updated using the corresponding function `update_waveforms()`.

        :param element: name of the element to get the waveforms from. Must be defined in the config.
        :param operation_name: name of the operation to get the waveforms from. Must be defined in the config.
        :param new_name: name of the new operation.
        """
        # Check inputs
        if element not in self.data["elements"].keys():
            raise KeyError(f"The element '{element}' in not defined in the config.")
        if operation_name not in self.data["elements"][element]["operations"].keys():
            raise KeyError(
                f"The operation '{operation_name}' in not defined in the config for the element '{element}'."
            )
        # Copy operation
        pulse_name = self.data["elements"][element]["operations"][operation_name]
        self.data["pulses"][element + "_" + new_name + self.pulse_suffix] = _deepcopy(self.data["pulses"][pulse_name])
        # Rename the copied operation
        self.data["elements"][element]["operations"][new_name] = element + "_" + new_name + self.pulse_suffix

    def update_waveforms(
        self,
        element: str,
        operation_name: str,
        wf: Tuple[List[float], List[float]] or Tuple[List[float]],
    ):
        """
        Update the waveforms from a specific operation and element.

        :param element: name of the element to get the waveforms from. Must be defined in the config.
        :param operation_name: name of the operation to get the waveforms from. Must be defined in the config.
        :param wf: Tuple containing the updated waveforms. (wf_single,) for a singleInput element and (wf_i, wf_q) for a mixInputs element.
        """
        # Check inputs
        if element not in self.data["elements"].keys():
            raise KeyError(f"The element '{element}' in not defined in the config.")
        if operation_name not in self.data["elements"][element]["operations"].keys():
            raise KeyError(
                f"The operation '{operation_name}' in not defined in the config for the element '{element}'."
            )
        if "mixInputs" not in self.data["elements"][element].keys() and len(wf) == 2:
            raise KeyError(f"Element '{element}' is not a mixInputs element.")
        if "singleInput" not in self.data["elements"][element].keys() and len(wf) == 1:
            raise KeyError(f"Element '{element}' is not a singleInput element.")
        if len(wf) == 2:
            if len(wf[0]) != len(wf[1]):
                raise ValueError("The 'I' and 'Q' waveforms must have the same length.")

        pulse_name = self.data["elements"][element]["operations"][operation_name]
        self.data["pulses"][pulse_name]["length"] = len(wf[0])
        # mixInputs element
        if len(wf) == 2:
            wf_name = element + "_" + operation_name + self.wf_suffix
            wf_i = wf[0]
            wf_q = wf[1]
            if len(wf_i) > 0:
                if wf_i[:-1] == wf_i[1:]:
                    self.data["waveforms"][wf_name + "_i"] = {
                        "type": "constant",
                        "sample": wf_i[0],
                    }
                else:
                    self.data["waveforms"][wf_name + "_i"] = {
                        "type": "arbitrary",
                        "samples": list(wf_i),
                    }

                if wf_q[:-1] == wf_q[1:]:
                    self.data["waveforms"][wf_name + "_q"] = {
                        "type": "constant",
                        "sample": wf_q[0],
                    }
                else:
                    self.data["waveforms"][wf_name + "_q"] = {
                        "type": "arbitrary",
                        "samples": list(wf_q),
                    }

            else:
                raise ValueError("The waveforms must have at least one point.")
            self.data["pulses"][pulse_name]["waveforms"] = {
                "I": wf_name + "_i",
                "Q": wf_name + "_q",
            }

        elif len(wf) == 1:
            wf_name = element + "_" + operation_name + "_single" + self.wf_suffix
            wf_s = wf[0]
            if len(wf_s) > 0:
                if wf_s[:-1] == wf_s[1:]:
                    self.data["waveforms"][wf_name] = {
                        "type": "constant",
                        "sample": wf_s[0],
                    }
                else:
                    self.data["waveforms"][wf_name] = {
                        "type": "arbitrary",
                        "samples": list(wf_s),
                    }
            else:
                raise ValueError("The waveform must have at least one point.")

            self.data["pulses"][pulse_name]["waveforms"] = {
                "single": wf_name,
            }



from qualang_tools.octave_tools.octave_tools import (
    get_calibration_parameters_from_db,
    set_correction_parameters_to_opx,
    get_correction_for_each_LO_and_IF,
    octave_calibration_tool,
)

__all__ = [
    "get_calibration_parameters_from_db",
    "set_correction_parameters_to_opx",
    "get_correction_for_each_LO_and_IF",
    "octave_calibration_tool",
]



from typing import Tuple, List

from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from qualang_tools.wirer.visualizer.layout import INSTRUMENT_FIGURE_DIMENSIONS, instrument_id_mapping


def _key(instrument_id: str, con: int):
    return f"{instrument_id}_{con}"


class InstrumentFigureManager:
    def __init__(self):
        self.figures = {}

    def get_ax(self, con: int, slot: int, instrument_id: str) -> Axes:
        instrument_id = instrument_id_mapping[instrument_id]
        key = _key(instrument_id, con)

        if key not in self.figures:
            if instrument_id == "OPX1000":
                fig, axs = self._make_opx1000_figure()
                self.figures[key] = {i + 1: ax for i, ax in enumerate(axs)}
                fig.suptitle(f"con{con} - {instrument_id} Wiring", fontweight="bold", fontsize=14)
            elif instrument_id == "OPX+":
                fig = self._make_opx_plus_figure()
                self.figures[key] = fig.axes[0]
                fig.suptitle(f"con{con} - {instrument_id} Wiring", fontweight="bold", fontsize=14)
            elif instrument_id == "Octave":
                fig = self._make_octave_figure()
                self.figures[key] = fig.axes[0]
                fig.suptitle(f"oct{con} - {instrument_id} Wiring", fontweight="bold", fontsize=14)
            else:
                fig = self._make_external_mixer_figure()
                self.figures[key] = fig.axes[0]
                fig.suptitle(f"Mixers {con} Wiring", fontweight="bold", fontsize=14)

        return self.figures[key][slot] if slot is not None else self.figures[key]

    @staticmethod
    def _make_opx1000_figure() -> Tuple[Figure, List[Axes]]:
        fig, axs = plt.subplots(
            nrows=1,
            ncols=8,
            figsize=(
                INSTRUMENT_FIGURE_DIMENSIONS["OPX1000"]["width"] * 2,
                INSTRUMENT_FIGURE_DIMENSIONS["OPX1000"]["height"] * 2,
            ),
        )
        for i, ax in enumerate(axs.flat):
            ax.set_ylim([0.15, 1.15])
            ax.set_xlim([0.15 / 8 * 3, 1.15 / 8 * 3])
            ax.set_facecolor("darkgrey")
            ax.set_xticks([])
            ax.set_yticks([])
            ax.set_xticklabels([])
            ax.set_yticklabels([])
            ax.set_aspect("equal")
        fig.subplots_adjust(wspace=0)

        return fig, axs

    @staticmethod
    def _make_opx_plus_figure() -> Figure:
        fig, ax = plt.subplots(
            1,
            1,
            figsize=(
                INSTRUMENT_FIGURE_DIMENSIONS["OPX+"]["width"] * 2,
                INSTRUMENT_FIGURE_DIMENSIONS["OPX+"]["height"] * 2,
            ),
        )
        ax.set_ylim([0.15 / 8 * 3, 1.15 / 8 * 3])
        ax.set_xlim([0.15 * 3, 1.15 * 3])
        ax.set_facecolor("darkgrey")
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.set_aspect("equal")

        return fig

    @classmethod
    def _make_octave_figure(cls) -> Figure:
        return cls._make_opx_plus_figure()

    @classmethod
    def _make_external_mixer_figure(cls) -> Figure:
        fig, ax = plt.subplots(
            1,
            1,
            figsize=(
                INSTRUMENT_FIGURE_DIMENSIONS["Mixers"]["width"] * 2,
                INSTRUMENT_FIGURE_DIMENSIONS["Mixers"]["height"] * 2,
            ),
        )
        ax.text(0.25, 0.25, "Up. RF", ha="center", va="center")
        ax.text(0.25, 0.65, "Up. Mkr", ha="center")
        ax.text(0.75, 0.25, "Down. RF", ha="center", va="center")
        ax.set_ylim([0.15, 1.15])
        # ax.set_xlim([0.15 / 8 * 3, 1.15 / 8 * 3])
        ax.set_facecolor("darkgrey")
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.set_aspect("equal")

        return fig



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import dash_bootstrap_components as dbc
from dash import Input, Output, State, MATCH, ALL, ALLSMALLER
from dash.exceptions import PreventUpdate
import pandas
import plotly.express as px
import traceback
import os
import importlib
from .app import app

from .config_editor import config_editor
from .component_editor import component_editor, get_component, class_editor_and_gui
from .upload import UPLOAD_DIRECTORY

__all__ = ["config_editor"]


@app.callback(
    [Output("update-success", "children")],
    [Input("update-button", "n_clicks")],
    [State("input-field", "value"), State("url", "pathname")],
)
def update_field(click_number, value, pathname):
    if isinstance(value, str):
        value = value.replace("\n", " ")
    if click_number == 0:
        return [""]
    success = config_editor(pathname, updated_value=value)
    if success == "OK":
        with open(os.path.join(UPLOAD_DIRECTORY, "config_edits.py"), "a") as myfile:
            if isinstance(value, str):
                myfile.write(f"config_editor('{pathname}', updated_value='{value}', configuration=configuration)\n")
            else:
                myfile.write(f"config_editor('{pathname}', updated_value={value}, configuration=configuration)\n")
        return [dbc.Alert(f"Updated {pathname} = {value}", color="success")]
    else:
        return [dbc.Alert(f"{success}", color="danger")]


@app.callback(
    [Output("update-success-tuple", "children")],
    [Input("update-tuple-button", "n_clicks")],
    [State("tuple-0", "value"), State("tuple-1", "value"), State("url", "pathname")],
    prevent_initial_call=True,
)
def update_tuple(click_number, value0, value1, pathname):
    if click_number == 0:
        return [""]
    success = config_editor(pathname, updated_value=(value0, value1))
    if success == "OK":
        with open(os.path.join(UPLOAD_DIRECTORY, "config_edits.py"), "a") as myfile:
            myfile.write(
                f"config_editor('{pathname}', updated_value=('{value0}',{value1}), configuration=configuration)\n"
            )

        return [dbc.Alert(f"Updated {pathname} = ('{value0}', {value1})", color="success")]
    else:
        return [dbc.Alert(f"{success}", color="danger")]


@app.callback(
    [Output("wf-view", "figure"), Output("wf-link", "href")],
    [Input("wf-dropdown", "value")],
    [State("url", "pathname"), State("wf-initial", "children")],
    prevent_initial_call=True,
)
def update_wf(value, pathname, initial_waveform):
    if value == initial_waveform:
        raise PreventUpdate

    success = config_editor(pathname, updated_value=value)
    if success == "OK":
        with open(os.path.join(UPLOAD_DIRECTORY, "config_edits.py"), "a") as myfile:
            if isinstance(value, str):
                myfile.write(f"config_editor('{pathname}', updated_value='{value}', config=configuration)\n")
            else:
                myfile.write(f"config_editor('{pathname}', updated_value={value}, config=configuration)\n")
        import config_edits
        import config_final

        importlib.reload(config_edits)
        importlib.reload(config_final)
        configuration = config_final.configuration

        wf = configuration["waveforms"][value]
        if wf["type"] == "arbitrary" or wf["type"] == "compressed":
            data = pandas.DataFrame(wf["samples"])
        else:
            data = pandas.DataFrame([wf["sample"]])
        fig = px.line(data, markers=True)
        fig.update_layout(showlegend=False)
        return [fig, f"/config/waveforms/{value}"]
    else:
        print("Error")
        print(success)
        raise PreventUpdate


@app.callback(
    [
        Output({"type": "port-list", "index": MATCH}, "value"),
        Output({"type": "port-list", "index": MATCH}, "options"),
    ],
    [Input({"type": "port-button", "index": MATCH}, "n_clicks")],
    [
        State({"type": "port-list", "index": MATCH}, "options"),
        State({"type": "port-list", "index": MATCH}, "value"),
        State({"type": "input-controller", "index": MATCH}, "value"),
        State({"type": "input-port", "index": MATCH}, "value"),
    ],
    prevent_initial_call=True,
)
def add_port_to_list(bclick, options, values, controller, port):
    if options is None:
        options = []
    if values is None:
        values = []
    controller = controller.split(",")
    controller_name = controller[0]
    controller_index = int(controller[1])
    newController = f"{controller_index}, {port}"
    if newController not in values:
        values.append(newController)
    newOption = {
        "label": f"{controller_name}, {port}",
        "value": f"{controller_index}, {port}",
    }
    if newOption not in options:
        options.append(newOption)
    return values, options


@app.callback(
    Output("component-status", "children"),
    Input("button-add-component", "n_clicks"),
    [
        State({"type": "component-input", "index": ALL}, "value"),
        State({"type": "port-list", "index": ALL}, "value"),
        State("component-input-add", "value"),
        State("qm-component-select", "value"),
    ],
    prevent_initial_call=True,
)
def add_component_click(bclick, inputs, ports, added_objects, selected_component):
    try:
        python_edit_command = component_editor(selected_component, inputs=[inputs, ports])
        if added_objects is not None:
            for index, item in enumerate(added_objects):
                if index == 0:
                    python_edit_command += f".add([{item}"
                else:
                    python_edit_command += f", {item}"
            if len(added_objects) > 0:
                python_edit_command += "])"
        with open(os.path.join(UPLOAD_DIRECTORY, "config_edits.py"), "a") as myfile:
            myfile.write(f"setup.add({python_edit_command})\n")

        return dbc.Alert(f"Component added successfuly!\n {python_edit_command}", color="success")
    except Exception:
        return dbc.Alert(traceback.format_exc(), color="danger")


@app.callback(
    [
        Output({"type": "optional-class-selection", "index": MATCH}, "value"),
        Output({"type": "optional-class-selection", "index": MATCH}, "options"),
        Output({"type": "optional-class-status", "index": MATCH}, "children"),
    ],
    [Input({"type": "button-add-option", "index": MATCH}, "n_clicks")],
    [
        State({"type": "optional-class-fieldcount", "index": MATCH}, "value"),
        State({"type": "optional-class-name", "index": MATCH}, "value"),
        State({"type": "optional-class-selection", "index": MATCH}, "value"),
        State({"type": "optional-class-selection", "index": MATCH}, "options"),
        State({"type": "optional-class-input", "index": ALLSMALLER}, "value"),
    ],
    prevent_intial_call=True,
)
def add_optional_input(
    button_click,
    fieldcount,
    class_name,
    current_added_items,
    current_options,
    class_input,
):
    fieldcount = int(fieldcount)
    from config_edits import setup

    component, component_class = get_component(class_name)

    try:
        python_command, forget, forget = class_editor_and_gui(
            component_class.__init__,
            setup,
            [],
            "",
            [class_input[-fieldcount:], []],
            [0, 0],
            input_types=["component-input", "port-list"],
        )

        python_command = component + "(" + python_command + ")"

        # add on current class list
        current_options.append({"label": python_command, "value": python_command})
        current_added_items.append(python_command)

        return (
            current_added_items,
            current_options,
            dbc.Alert(f"Component added successfuly!\n {python_command}", color="success"),
        )
    except Exception:
        raise PreventUpdate
        # for debugging purposes comment previous line and let this be seen
        return (
            current_added_items,
            current_options,
            dbc.Alert(traceback.format_exc(), color="danger"),
        )


@app.callback(
    [Output("component-input-add", "value"), Output("component-input-add", "options")],
    [Input({"type": "optional-class-selection", "index": ALL}, "value")],
    [State({"type": "optional-class-selection", "index": ALL}, "options")],
    prevent_intial_call=True,
)
def collect_all_added_objects(listed_objects, possible_options):
    all_objects = []
    all_options = []
    for l in listed_objects:
        all_objects += l
    for o in possible_options:
        all_options += o

    return all_objects, all_options



import importlib
from dash import dcc
import dash_bootstrap_components as dbc
from dash import html
import plotly.express as px
import json
import pprint
import numpy as np
import pandas
import html as htmlpy
from .component_editor import editor_of_quantum_machine_elements
from .upload import UPLOAD_DIRECTORY
import os
import urllib.request

__all__ = ["config_editor"]

config_structure = {}

if not os.path.exists(UPLOAD_DIRECTORY):
    os.makedirs(UPLOAD_DIRECTORY)

try:
    # try updating config schema
    print("\tDownloading latest config schema...")
    with urllib.request.urlopen("https://qm-docs.qualang.io/qm_config_spec.json") as url:
        config_structure = json.loads(url.read().decode())
    print("\tDONE")
except Exception:
    print("Cannot download. Using the local copy of the schema.")
    with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "qua1_openapi.json")) as file:
        config_structure = json.load(file)


def getDefinition(path):
    key = path.split("/")[2]
    return config_structure["definitions"][key]


def config_editor(pathname, updated_value=None, configuration=None):
    location = pathname.split("/")[2:]

    if configuration is None:
        import config_edits
        import config_final

        importlib.reload(config_edits)
        importlib.reload(config_final)
        configuration = config_final.configuration

    a = configuration

    breadcrumb_items = [{"label": "Config", "href": "/config", "external_link": True}]
    url = "/config"
    selected = ""
    documentation = config_structure["definitions"]["QmConfig"]["properties"]
    waveform_view = False
    pulse_view = False
    controlArguments = ""

    # dive into requested location in the config
    for l in location:
        if (l == "waveforms") and (len(location) > 1 and location[0] != "waveforms"):
            waveform_view = True
        if l == "operations":
            pulse_view = True
        if l != "":
            if l[0] == "[":
                controlArguments = l
                break

            try:
                l = int(l)
            except ValueError:
                pass
            # print(l)
            # print(a)

            if isinstance(a[l], dict) or (isinstance(a[l], list) and len(a[l]) > 0 and isinstance(a[l][0], dict)):
                a = a[l]
                if l in documentation:
                    documentation = documentation[l]
                else:
                    if "properties" in documentation and l in documentation["properties"]:
                        documentation = documentation["properties"][l]
                        if "$ref" in documentation:
                            documentation = getDefinition(documentation["$ref"])
                    elif "additionalProperties" in documentation:
                        if "$ref" in documentation["additionalProperties"]:
                            documentation = getDefinition(documentation["additionalProperties"]["$ref"])
                            # print(documentation)
                        elif "oneOf" in documentation["additionalProperties"]:
                            for option in documentation["additionalProperties"]["oneOf"]:
                                # print("\n")
                                option = getDefinition(option["$ref"])
                                try:
                                    # print(option)
                                    # print(a["type"])
                                    if option["properties"]["type"]["description"].find(a["type"]) != -1:
                                        documentation = option
                                except KeyError:
                                    pass

                    else:
                        documentation = {}
                url += "/" + str(l)
                breadcrumb_items.append(
                    {
                        "label": l,
                        "href": url,
                        "external_link": True,
                    }
                )
            elif pulse_view and (isinstance(a[l], str)):
                pulse_view = False
                selected_pulse = a[l]
                a = configuration["pulses"][selected_pulse]
                url += "/" + l
                breadcrumb_items.append(
                    {
                        "label": ("%s -> %s" % (l, selected_pulse)),
                        "href": url,
                        "external_link": True,
                    }
                )
            else:
                selected = l

    breadcrumb_items[-1]["active"] = True
    input_dict = a

    # print("selected ",selected)
    # print(input_dict)
    # print(controlArguments)

    list_items = []
    for key, value in input_dict.items():
        val = " = "
        className = "text-white"
        if key != selected:
            className = "text-muted"
            if isinstance(value, str):
                val = f" = '{value}'"
            elif isinstance(value, float):
                val = f" = {value}"
            elif isinstance(value, int):
                val = f" = {value}"
            else:
                val = " ..."
        list_items.append(
            dbc.ListGroupItem(
                html.Div(
                    [
                        html.H5(str(key), className="mb-1"),
                        html.Small(
                            val,
                            className=className,
                        ),
                    ],
                    className="d-flex w-100 justify-content-between",
                ),
                href=url + "/" + str(key),
                action=True,
                active=key == selected,
            )
        )
    list_group = dbc.ListGroup(list_items)
    details = []
    docs = []
    # print(documentation)

    # collect relevant documentation
    if "description" in documentation:
        if "title" in documentation:
            docs.append(
                html.H4(
                    "%s: %s" % (breadcrumb_items[-1]["label"], documentation["title"]),
                    className="mb-1",
                )
            )
        docs.append(html.Div(documentation["description"]))
        if "properties" in documentation:
            list_items = []
            for key, value in documentation["properties"].items():
                item = []
                item.append(html.Strong("%s" % key))
                if "$ref" in value:
                    # print(value)
                    value = getDefinition(value["$ref"])
                    # print(value)
                if "type" in value:
                    item.append(html.Span(" (%s)" % value["type"]))
                if "default" in value:
                    item.append(html.Span(", default '%s'" % value["default"]))
                if "description" in value:
                    item.append(html.Span(": %s" % value["description"]))
                list_items.append(dbc.ListGroupItem(item))
            docs.append(dbc.ListGroup(list_items))

    # show selected values
    if selected != "":
        if isinstance(input_dict[selected], list) or isinstance(input_dict[selected], np.ndarray):
            if isinstance(input_dict[selected], np.ndarray):
                input_dict[selected] = input_dict[selected].tolist()

            # print(url)
            if updated_value is not None:
                updated_value = updated_value.replace("[", "")
                updated_value = updated_value.replace("]", "")
                input_dict[selected] = np.fromstring(updated_value, sep=",")
                return "OK"
            if controlArguments.find("view=list") != -1:

                details.append(dcc.Link("view as plot", href=url + f"/{selected}/[view=plot]"))
                # show list
                details.append(
                    dbc.InputGroup(
                        [
                            dbc.Textarea(
                                id="input-field",
                                className="mb-3",
                                placeholder="array",
                                value=htmlpy.escape(pprint.pformat(input_dict[selected])),
                            ),
                            dbc.Button("Update", id="update-button", n_clicks=0),
                        ]
                    )
                )
                details.append(html.Div(id="update-success"))
            else:
                details.append(dcc.Link("view as list", href=url + f"/{selected}/[view=list]"))
                # show plot
                data = pandas.DataFrame(input_dict[selected])
                fig = px.line(data, markers=True)
                fig.update_layout(showlegend=False)
                details.append(dcc.Graph(id="graph", config={"displayModeBar": False}, figure=fig))
        else:
            if isinstance(input_dict[selected], float):
                if updated_value is not None:
                    # do checks
                    input_dict[selected] = updated_value
                    return "OK"
                # else show edit thing
                details.append(
                    dbc.InputGroup(
                        [
                            dbc.Input(
                                id="input-field",
                                placeholder="Enter value...",
                                type="number",
                                value=("%.8e" % input_dict[selected]),
                            ),
                            dbc.Button("Update", id="update-button", n_clicks=0),
                        ]
                    )
                )
                details.append(html.Div(id="update-success"))
            else:
                if waveform_view and isinstance(input_dict[selected], str):
                    options = []
                    for wfoption in configuration["waveforms"].keys():
                        options.append({"label": f"{wfoption}", "value": f"{wfoption}"})
                    details.append(
                        dcc.Dropdown(
                            id="wf-dropdown",
                            options=options,
                            searchable=True,
                            value=input_dict[selected],
                            clearable=False,
                        )
                    )
                    details.append(html.Span(input_dict[selected], id="wf-initial", hidden=True))
                    if updated_value is not None:
                        input_dict[selected] = updated_value
                        return "OK"
                    wf = configuration["waveforms"][input_dict[selected]]
                    if wf["type"] == "arbitrary" or wf["type"] == "compressed":
                        data = pandas.DataFrame(wf["samples"])
                    else:
                        data = pandas.DataFrame([wf["sample"]])
                    fig = px.line(data, markers=True)
                    fig.update_layout(showlegend=False)
                    details.append(dcc.Graph(id="wf-view", config={"displayModeBar": False}, figure=fig))

                    details.append(
                        dcc.Link(
                            "open this waveform",
                            href=("/config/waveforms/%s" % input_dict[selected]),
                            id="wf-link",
                        )
                    )
                else:
                    if isinstance(input_dict[selected], tuple):
                        if updated_value is not None and isinstance(updated_value, tuple):
                            input_dict[selected] = updated_value
                            return "OK"
                        details.append(
                            dbc.InputGroup(
                                [
                                    dbc.InputGroupText("("),
                                    dbc.Input(
                                        id=("tuple-0"),
                                        placeholder="Enter value...",
                                        type="text",
                                        value=("%s" % input_dict[selected][0]),
                                    ),
                                    dbc.InputGroupText(","),
                                    dbc.Input(
                                        id=("tuple-1"),
                                        placeholder="Enter value...",
                                        type="number",
                                        value=("%s" % input_dict[selected][1]),
                                    ),
                                    dbc.InputGroupText(")"),
                                    dbc.Button("Update", id="update-tuple-button", n_clicks=0),
                                ],
                                className="mb-3",
                            )
                        )
                        details.append(details.append(html.Div(id="update-success-tuple")))
                    else:
                        if updated_value is not None:
                            # do checks
                            input_dict[selected] = updated_value
                            return "OK"
                        details.append(
                            dbc.InputGroup(
                                [
                                    dbc.Input(
                                        id="input-field",
                                        placeholder="Enter value...",
                                        type="text",
                                        value=("%s" % input_dict[selected]),
                                    ),
                                    dbc.Button("Update", id="update-button", n_clicks=0),
                                ]
                            )
                        )
                        details.append(html.Div(id="update-success"))

    # add everything to final page
    return html.Div(
        [
            dbc.Breadcrumb(
                items=breadcrumb_items,
            ),
            dbc.Row(
                [
                    dbc.Col(list_group, width=3),
                    dbc.Col(details, width=6),
                    dbc.Col(docs, width=3),
                ]
            ),
            editor_of_quantum_machine_elements(),
        ]
    )



from qualang_tools.simulator import create_simulator_controller_connections
import warnings

__all__ = [
    "create_simulator_controller_connections",
]

warnings.warn(
    "The 'create_simulator_controller_connections' function has been moved to 'qualang_tools.simulator'.", FutureWarning
)
warnings.warn(
    "The 'create_simulator_controller_connections' function has been moved to 'qualang_tools.simulator'.",
    DeprecationWarning,
)



from qm.qua import *


def assign_variables_to_element(element, *variables):
    """
    Forces the given variables to be used by the given element thread. Useful as a workaround for when the compiler
    wrongly assigns variables which can cause gaps.
    To be used at the beginning of a program, will add a 16ns wait to the given element. Use an `align()` if needed.

    Example::

        >>> with program() as program_name:
        >>>     a = declare(int)
        >>>     b = declare(fixed)
        >>>     assign_variables_to_element('resonator', a, b)
        >>>     align()
        >>>     ...

    """
    _exp = Cast.to_int(variables[0])
    for variable in variables[1:]:
        _exp += Cast.to_int(variable)
    wait(4 + 0 * _exp, element)



from dataclasses import asdict
from typing import Type

from qualang_tools.wirer.instruments.instrument_channel import (
    InstrumentChannel,
    InstrumentChannelLfFemInput,
    InstrumentChannelLfFemOutput,
    InstrumentChannelMwFemInput,
    InstrumentChannelMwFemOutput,
    InstrumentChannelOpxPlusInput,
    InstrumentChannelOpxPlusOutput,
    AnyInstrumentChannel,
    InstrumentChannelLfFemDigitalOutput,
    InstrumentChannelMwFemDigitalOutput,
    InstrumentChannelOpxPlusDigitalOutput,
)

CHANNELS_OPX_PLUS = [
    InstrumentChannelOpxPlusInput,
    InstrumentChannelOpxPlusOutput,
    InstrumentChannelOpxPlusDigitalOutput,
]

CHANNELS_OPX_1000 = [
    InstrumentChannelLfFemInput,
    InstrumentChannelLfFemOutput,
    InstrumentChannelLfFemDigitalOutput,
    InstrumentChannelMwFemInput,
    InstrumentChannelMwFemOutput,
    InstrumentChannelMwFemDigitalOutput,
]


class InstrumentChannels:
    """
    Collection of "stack" data-structures organized by channel type.
    Each entry contains a list of instrument channels. This wrapper
    can be interacted with as if it were the underlying stack dictionary.
    """

    def __init__(self):
        self.stack = {}

    def check_if_already_occupied(self, channel: AnyInstrumentChannel):
        for channel_type in self.stack:
            for existing_channel in self.stack[channel_type]:
                if (
                    channel.con == existing_channel.con
                    and channel.slot == existing_channel.slot
                    and channel.port == existing_channel.port
                    and channel.io_type == existing_channel.io_type
                    and channel.signal_type == existing_channel.signal_type
                ):
                    if channel.slot is None:
                        if type(channel) is not type(existing_channel):
                            pass
                        else:
                            raise ValueError(
                                f"{channel.io_type} channel on con{channel.con}, "
                                f"port {channel.port} is already occupied."
                            )
                    else:
                        raise ValueError(
                            f"{channel.io_type} channel on con{channel.con}, "
                            f"slot {channel.slot}, port {channel.port} is already occupied."
                        )

    def check_if_mixing_opx_1000_and_opx_plus(self, channel: InstrumentChannel):
        if type(channel) in CHANNELS_OPX_1000:
            for channel_type in CHANNELS_OPX_PLUS:
                if channel_type in self.stack:
                    raise ValueError("Can't add an FEM to a setup with an OPX+.")
        elif type(channel) in CHANNELS_OPX_PLUS:
            for channel_type in CHANNELS_OPX_1000:
                if channel_type in self.stack:
                    raise ValueError("Can't add an OPX+ to a setup with an OPX1000 FEM.")

    def insert(self, pos: int, channel: InstrumentChannel):
        self.check_if_already_occupied(channel)
        self.check_if_mixing_opx_1000_and_opx_plus(channel)

        channel_type = type(channel)
        if channel_type not in self.stack:
            self.stack[channel_type] = []

        self.stack[channel_type].insert(pos, channel)

    def add(self, channel: InstrumentChannel):
        self.check_if_already_occupied(channel)
        self.check_if_mixing_opx_1000_and_opx_plus(channel)

        channel_type = type(channel)
        if channel_type not in self.stack:
            self.stack[channel_type] = []

        self.stack[channel_type].append(channel)

    def pop(self, channel: Type[InstrumentChannel]):
        return self.stack[channel].pop(0)

    def remove(self, channel: InstrumentChannel):
        return self.stack[type(channel)].remove(channel)

    def get(self, key: InstrumentChannel, fallback=None):
        return self.stack.get(key, fallback)

    def __getitem__(self, item):
        return self.stack[item]

    def __iter__(self):
        return iter(self.stack)

    def __contains__(self, item: InstrumentChannel):
        for channel_type in self.stack:
            if asdict(item) in [asdict(channel) for channel in self.stack[channel_type]]:
                return True
        return False

    def items(self):
        return self.stack.items()



from typing import List

import matplotlib.pyplot as plt

from qualang_tools.wirer.connectivity.wiring_spec import WiringLineType
from qualang_tools.wirer.instruments.instrument_channels import InstrumentChannels
from qualang_tools.wirer.visualizer.instrument_figure_manager import InstrumentFigureManager
from qualang_tools.wirer.visualizer.port_annotation import PortAnnotation


def invert_qubit_dict(qubit_dict) -> dict:
    inverted_dict = {}
    for qubit_ref, element in qubit_dict.items():
        for channel_type, channels in element.channels.items():
            for channel in channels:
                key = (
                    channel.con,
                    channel.slot,
                    channel.port,
                    channel.io_type,
                    channel.signal_type,
                    channel.instrument_id,
                    channel_type,
                )
                if key not in inverted_dict:
                    inverted_dict[key] = []
                annotation = f"{element.id}.{channel_type if isinstance(channel_type, str) else channel_type.value}"
                inverted_dict[key].append((annotation, channel))
    return inverted_dict


def make_annotations(inverted_dict: dict) -> List[PortAnnotation]:
    annotations = []
    for key, values in inverted_dict.items():
        con, slot, port, io_type, signal_type, instrument_id, channel_type = key
        labels = [value[0] for value in values]
        color = get_color_for_line_type(channel_type)
        annotations.append(PortAnnotation(labels, color, con, slot, port, io_type, signal_type, instrument_id))

    return annotations


def merge_annotations_on_same_channel(annotations: List[PortAnnotation]) -> List[PortAnnotation]:
    annotations_by_channel = dict()
    for annotation in annotations:
        channel_address = (
            annotation.con,
            annotation.slot,
            annotation.port,
            annotation.io_type,
            annotation.signal_type,
            annotation.instrument_id,
        )
        annotations_at_channel_address = annotations_by_channel.get(channel_address, [])
        annotations_by_channel[channel_address] = annotations_at_channel_address + [annotation]

    merged_annotations = []
    for annotations_at_channel_address in annotations_by_channel.values():
        base_annotation = annotations_at_channel_address[0]
        for annotation in annotations_at_channel_address[1:]:
            base_annotation.labels += annotation.labels
        base_annotation.labels = sorted(base_annotation.labels)
        merged_annotations.append(base_annotation)

    return merged_annotations


def make_unused_channel_annotations(available_channels: InstrumentChannels) -> List[PortAnnotation]:
    annotations = []
    for _, channel_list in available_channels.stack.items():
        for channel in channel_list:
            annotations.append(
                PortAnnotation(
                    [""],
                    "white",
                    channel.con,
                    channel.slot,
                    channel.port,
                    channel.io_type,
                    channel.signal_type,
                    channel.instrument_id,
                )
            )

    return annotations


def get_color_for_line_type(line_type) -> str:
    color_map = {
        WiringLineType.FLUX: "powderblue",
        WiringLineType.CHARGE: "lavender",
        WiringLineType.RESONATOR: "peachpuff",
        WiringLineType.DRIVE: "lemonchiffon",
        WiringLineType.COUPLER: "thistle",
    }
    return color_map.get(line_type, "beige")


def draw_annotations(manager: InstrumentFigureManager, annotations: List[PortAnnotation]):
    for annotation in annotations:
        ax = manager.get_ax(annotation.con, annotation.slot, annotation.instrument_id)
        annotation.draw(ax)
        annotation.title_axes(ax)


def visualize(qubit_dict, available_channels=None):
    # Invert the qubit dictionary for easier annotation processing
    inverted_dict = invert_qubit_dict(qubit_dict)

    # Prepare annotations and labels
    annotations = make_annotations(inverted_dict)
    annotations = merge_annotations_on_same_channel(annotations)

    # Manage figures and draw
    manager = InstrumentFigureManager()

    if available_channels is not None:
        available_channel_annotations = make_unused_channel_annotations(available_channels)
        draw_annotations(manager, available_channel_annotations)

    draw_annotations(manager, annotations)

    plt.show()
